/*
  BiologicalCell.m

  Implementation of biological cell

  Copyright (C) 2005 Scott Christley

  Author: Scott Christley <schristl@nd.edu>
  Date: 2005
   
  This file is part of LimbChondrogenesis.
*/

#import "BiologicalCell.h"
#import "MultiScale2DGrid.h"
#import "Grid2DArray.h"

#include "internal.h"

/* check_overlap can check if the cells overlap.  */
/* If they do, it returns NO and we can not accept the movement. */
/* If they don't, it returns YES, and we can accept the movement. */  
BOOL check_overlap(void *matrix, unsigned int width, unsigned int height, unsigned int *Anew,
		   unsigned int *Bnew, unsigned int *A, unsigned int *B, unsigned int size)
{
  int i, j;
  int Diff[size];  //Mark of the difference between A,B and Anew,Bnew.
  id (*grid)[height][width] = matrix;

  for (i = 0; i < size; i++)   Diff[i]=1;

  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      if (Anew[i] == A[j] && Bnew[i] == B[j]) {
	Diff[i] = 0;
      }
    }
  }
  
  for (i = 0; i < size; i++) {
    if (Diff[i] == 1) {
      if ((*grid)[Bnew[i]][Anew[i]] != nil) {
	return NO;
      }
    }
  }

  return YES;
}


/* we accept the movement and use rewrite_cell to record to new position of the pixels  */
/* in the grid and their coordinates in A and B.  */
void rewrite_cell(void *matrix, unsigned int width, unsigned int height, unsigned int *Anew,
		  unsigned int *Bnew, unsigned int *A, unsigned int *B, unsigned int size, id aCell)
{
  int i;
  id (*grid)[height][width] = matrix;

  for (i = 0; i < size; i++) {
    if ((*grid)[B[i]][A[i]] != aCell) printf("wrong cell %d %d!\n", B[i], A[i]);
    (*grid)[B[i]][A[i]] = nil;
  }

  for (i = 0; i < size; i++) {
    if ((*grid)[Bnew[i]][Anew[i]] != nil) printf("existing cell %d %d!\n", Bnew[i], Anew[i]);
    (*grid)[Bnew[i]][Anew[i]] = aCell;
  }

  for (i = 0; i < size; i++) {
    A[i] = Anew[i];
    B[i] = Bnew[i];
  }
}

@implementation BiologicalCell

- initWithColony: (CellColony *)aColony ofSize: (unsigned int)aSize
	     atX: (unsigned int)x atY: (unsigned int)y ofType: (int)aType
{
  [super init];

  cellColony = aColony;
  cellSize = aSize;
  cellType = aType;
  spatialExtent[0] = (unsigned int *)malloc(sizeof(unsigned int) * cellSize);
  spatialExtent[1] = (unsigned int *)malloc(sizeof(unsigned int) * cellSize);
  fibroProduction = 0;
  actMemoryProduce = 0;
  activatorSum = 0;

  [self buildCellAtX: (unsigned int)x atY: (unsigned int)y];

  return self;
}

- (void)dealloc
{
  free(spatialExtent[0]);
  free(spatialExtent[1]);

  [super dealloc];
}

- (int)cellType { return cellType; }
- (void)setCellType: (int)aType { cellType = aType; }
- (int)intValue { return cellType + 1; }
- (unsigned int)cellSize { return cellSize; }
- (unsigned int *)xSpatialExtent { return spatialExtent[0]; }
- (unsigned int *)ySpatialExtent { return spatialExtent[1]; }

- (int)fibroProduction { return fibroProduction; }
- (void)incrementFibroProduction: (int)aNum { fibroProduction += aNum; }

- (void)recordActivatorProduce: (int)aNum { actMemoryProduce += aNum; }
- (int)activatorProduce { return actMemoryProduce; }

- (int)activatorSum { return activatorSum; }
- (void)setActivatorSum: (int)aNum {activatorSum = aNum; }

- (void)setStuckOnFibronectin: (BOOL)aFlag { stuckOnFibronectin = aFlag; }
- (BOOL)stuckOnFibronectin { return stuckOnFibronectin; }

- (int)activatorThreshold { return activatorThreshold; }
- (void)setActivatorThreshold: (int)mean withSD: (int)sd
{
  activatorThreshold = (int)normal_dist_rand(mean, sd);
}

- (void)buildCellAtX: (unsigned int)x atY: (unsigned int)y
{
  NSDictionary *d = [cellColony spatialGrid];
  NSNumber *n = [d objectForKey: @"width"];
  int width = [n intValue];
  n = [d objectForKey: @"height"];
  int height = [n intValue];
  Grid2DArray *m = [d objectForKey: @"matrix"];
  id (*grid)[height][width] = [m getMatrix];

  // base 2x2 nucleus
  //if ((*grid)[y][x] != nil) printf("problem: %d %d\n", x, y);
  (*grid)[y][x] = self;
  spatialExtent[0][0] = x;
  spatialExtent[1][0] = y;
  //if ((*grid)[y+1][x] != nil) printf("problem: %d %d\n", x, y+1);
  (*grid)[y + 1][x] = self;
  spatialExtent[0][1] = x;
  spatialExtent[1][1] = y + 1;
  //if ((*grid)[y][x+1] != nil) printf("problem: %d %d\n", x+1, y);
  (*grid)[y][x + 1] = self;
  spatialExtent[0][2] = x + 1;
  spatialExtent[1][2] = y;
  //if ((*grid)[y+1][x+1] != nil) printf("problem: %d %d\n", x+1, y+1);
  (*grid)[y + 1][x + 1] = self;
  spatialExtent[0][3] = x + 1;
  spatialExtent[1][3] = y + 1;

  // 3 protrusion pixels on 3 sides of the 2x2 nucleus

  // determine side with no protusion
  int unaccept = (int) (RAND_NUM * 4);

  int C[8],D[8];
  int pixels_pro;
  int i, temp0, temp1, temp2;
  double r2;

  // possible places to put protusion
  // two positions on each protusion side
  C[0] = x+1;
  D[0] = y+2;
  C[1] = x;
  D[1] = y+2;
  C[2] = x-1;
  D[2] = y+1;
  C[3] = x-1;
  D[3] = y;
  C[4] = x;
  D[4] = y-1;
  C[5] = x+1;
  D[5] = y-1;
  C[6] = x+2;
  D[6] = y;
  C[7] = x+2;
  D[7] = y+1;  
  pixels_pro = 0;

  // boundary condition
  for (i = 0; i < 8; i++) {
    if (C[i] == -1)         C[i] = width-1;
    if (C[i] == width)      C[i] = 0;
    if (C[i] == width+1)    C[i] = 1;
    if (D[i] == -1)         D[i] = height-1;
    if (D[i] == height)     D[i] = 0;
    if (D[i] == height+1)   D[i] = 1;
  }

  for (i = 0; i < 4; i++) {
    if (i == unaccept) continue;

    temp0 = i*2;
    temp1 = i*2+1;
    temp2 = pixels_pro+4;

    // if another cell has the one position
    // then take the other position
    if ((*grid)[D[temp0]][C[temp0]] != nil) {
      //if ((*grid)[D[temp1]][C[temp1]] != nil) printf("problem: %d %d\n", D[temp1], C[temp1]);
      (*grid)[D[temp1]][C[temp1]] = self;
      spatialExtent[0][temp2] = C[temp1];
      spatialExtent[1][temp2] = D[temp1];
      pixels_pro++;
      continue;
    }

    // if another cell has the one position
    // then take the other position
    if ((*grid)[D[temp1]][C[temp1]] != nil) {
      //if ((*grid)[D[temp0]][C[temp0]] != nil) printf("problem: %d %d\n", D[temp0], C[temp0]);
      (*grid)[D[temp0]][C[temp0]] = self;
      spatialExtent[0][temp2] = C[temp0];
      spatialExtent[1][temp2] = D[temp0];
      pixels_pro++;
      continue;
    }

    // randomly pick one of the two positions
    r2 = RAND_NUM;
    if (r2 < 1/2.0) {
      //if ((*grid)[D[temp0]][C[temp0]] != nil) printf("problem: %d %d\n", D[temp0], C[temp0]);
      (*grid)[D[temp0]][C[temp0]] = self;
      spatialExtent[0][temp2] = C[temp0];
      spatialExtent[1][temp2] = D[temp0];
    } else {
      //if ((*grid)[D[temp1]][C[temp1]] != nil) printf("problem: %d %d\n", D[temp1], C[temp1]);
      (*grid)[D[temp1]][C[temp1]] = self;
      spatialExtent[0][temp2] = C[temp1];
      spatialExtent[1][temp2] = D[temp1];
    }
    pixels_pro++;
  }
}

- (void)changeShape
{
  NSDictionary *d = [cellColony spatialGrid];
  NSNumber *n = [d objectForKey: @"width"];
  int width = [n intValue];
  n = [d objectForKey: @"height"];
  int height = [n intValue];
  n = [d objectForKey: @"boundaryCondition"];
  int bCond = [n intValue];
  Grid2DArray *m = [d objectForKey: @"matrix"];
  id (*grid)[height][width] = [m getMatrix];

  int i,j;
  id gridtest[4][4];
  int x0,y0;
  int side1[4],side2[4],sum_side1;
  int C[8],D[8],E[8],F[8];
  int pro_pixels;
  double P;
  double r1,r2;

  // corner of nucleus
  x0 = spatialExtent[0][0];
  y0 = spatialExtent[1][0];

  for (i = 0; i < 4; i++) {
    side1[i]=0;
    side2[i]=0;
  }

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      gridtest[i][j]=0;

  // possible protusions
  C[0] = x0+1;
  D[0] = y0+2;
  C[1] = x0;
  D[1] = y0+2;
  C[2] = x0-1;
  D[2] = y0+1;
  C[3] = x0-1;
  D[3] = y0;
  C[4] = x0;
  D[4] = y0-1;
  C[5] = x0+1;
  D[5] = y0-1;
  C[6] = x0+2;
  D[6] = y0;
  C[7] = x0+2;
  D[7] = y0+1;  

  if (bCond == PERIODIC_BOUNDARY) {
    // Periodic boundary condition
    for (i = 0; i < 8; i++) {
      if (C[i] == -1)         C[i] = width-1;
      if (C[i] == width)      C[i] = 0;
      if (C[i] == width+1)    C[i] = 1;
      if (D[i] == -1)         D[i] = height-1;
      if (D[i] == height)     D[i] = 0;
      if (D[i] == height+1)   D[i] = 1;
    }
  } else {
    // no flux boundary condition

    // hard to calculate, so if one of the protusions crosses the boundary
    // then just give up and disallow shape change
    for (i = 0; i < 8; i++) {
      if (C[i] == -1)         return;
      if (C[i] == width)      return;
      if (C[i] == width+1)    return;
      if (D[i] == -1)         return;
      if (D[i] == height)     return;
      if (D[i] == height+1)   return;
    }
  }

  E[0]=2;      
  F[0]=3;
  E[1]=1;
  F[1]=3;
  E[2]=0;
  F[2]=2;
  E[3]=0;
  F[3]=1;
  E[4]=1;
  F[4]=0;
  E[5]=2;
  F[5]=0;
  E[6]=3;
  F[6]=1;
  E[7]=3;
  F[7]=2; 

  // clear protusion areas
  for (i = 4; i < cellSize; i++)
    (*grid)[spatialExtent[1][i]][spatialExtent[0][i]] = nil;

  for (i = 0; i < 8; i++)
    gridtest[F[i]][E[i]] = (*grid)[D[i]][C[i]];

  // determine whether none, one, or two protusion areas available on each side
  for (i = 0; i < 4; i++) {
    if (gridtest[F[i*2]][E[i*2]] == nil || gridtest[F[i*2+1]][E[i*2+1]] == nil) {
      side1[i]=1;
      side2[i]=1;
    }
    if (gridtest[F[i*2]][E[i*2]] == nil && gridtest[F[i*2+1]][E[i*2+1]] == nil)  
      side2[i]=2;
  }

  sum_side1 = 0;
  for (i = 0; i < 4; i++) {
    sum_side1 = sum_side1 + side1[i];
  }
  if (sum_side1 < (cellSize - 4)) {
    printf("Debug1: %d!\n", sum_side1);
  }

  // random exclude sides
  // thus determine which ones will have protusions
  P = 0.25;
  pro_pixels = 0;
  while (sum_side1 > (cellSize - 4)) {
    r1 = RAND_NUM;
    for (i = 0; i < 4; i++)
      if ((r1 >= i*P) && (r1 < (i+1)*P))   side1[i]=0;

    sum_side1 = 0;
    for (i = 0; i < 4; i++) {
      sum_side1 = sum_side1 + side1[i];
    }
  }

  for (i = 0; i < 4; i++) {
    if (side1[i] == 0)   continue;
    if (side2[i] == 2) {
      r2 = RAND_NUM;
      /*  printf("r2=%lf\n",r2); */
      if (r2 < 0.5) {
	(*grid)[D[i*2]][C[i*2]] = self;
	spatialExtent[0][pro_pixels+4] = C[i*2];
	spatialExtent[1][pro_pixels+4] = D[i*2];
      } else {
	// r2 >= 0.5
	(*grid)[D[i*2+1]][C[i*2+1]] = self;
	spatialExtent[0][pro_pixels+4] = C[i*2+1];
	spatialExtent[1][pro_pixels+4] = D[i*2+1];
      } 
    } else {
      //  side2[i]==1
      if (gridtest[F[i*2]][E[i*2]] == nil) {
	(*grid)[D[i*2]][C[i*2]] = self;
	spatialExtent[0][pro_pixels+4] = C[i*2];
	spatialExtent[1][pro_pixels+4] = D[i*2];
      } else {
	(*grid)[D[i*2+1]][C[i*2+1]] = self;
	spatialExtent[0][pro_pixels+4] = C[i*2+1];
	spatialExtent[1][pro_pixels+4] = D[i*2+1];
      } 
    }
    pro_pixels++;
  }
}

- (void)moveWithProbability: (double)aProb
{
  NSDictionary *d = [cellColony spatialGrid];
  NSNumber *n = [d objectForKey: @"width"];
  int width = [n intValue];
  n = [d objectForKey: @"height"];
  int height = [n intValue];
  n = [d objectForKey: @"boundaryCondition"];
  int bCond = [n intValue];
  Grid2DArray *m = [d objectForKey: @"matrix"];
  id (*grid)[height][width] = [m getMatrix];

  double r;
  int i;
  unsigned int Anew[cellSize], Bnew[cellSize];

  // probabilities to move in one of four directions
  double Pmove = aProb / 4;
  double Pdown = Pmove;
  double Pleft = Pdown + Pmove;
  double Pup = Pleft + Pmove;
  double Pright = Pup + Pmove;

  r = RAND_NUM;
  if (r < Pdown) {
    // Move down.
    //printf("Move down\n");
    for (i = 0; i < cellSize; i++) {
      Anew[i] = spatialExtent[0][i];
      Bnew[i] = spatialExtent[1][i] + 1;
      if (Bnew[i] == height) {
	if (bCond == PERIODIC_BOUNDARY)
	  Bnew[i] = 0;    // periodic boundary condition
	else
	  return;  // no flux boundary, cannot move down
      }
    }
  } else if (r >= Pdown && r < Pleft) {
    // Move left.
    //printf("Move left\n");
    for (i = 0; i < cellSize; i++) {
      Anew[i] = spatialExtent[0][i] - 1;
      Bnew[i] = spatialExtent[1][i];
      if (Anew[i] == -1) {
	if (bCond == PERIODIC_BOUNDARY)
	  Anew[i] = width - 1;    // periodic boundary condition
	else
	  return;   // no flux boundary, cannot move left
      }
    }
  } else if (r >= Pleft && r < Pup) {
    // Move up.
    //printf("Move up\n");
    for (i = 0; i < cellSize; i++) {
      Anew[i] = spatialExtent[0][i];
      Bnew[i] = spatialExtent[1][i] - 1;
      if (Bnew[i] == -1) {
	if (bCond == PERIODIC_BOUNDARY)
	  Bnew[i] = height - 1;    // periodic boundary condition
	else
	  return;  // no flux boundary, cannot move up
      }
    }
  } else if (r >= Pup && r < Pright) {
    // Move right.
    //printf("Move right\n");
    for (i = 0; i < cellSize; i++) {
      Anew[i] = spatialExtent[0][i] + 1;
      Bnew[i] = spatialExtent[1][i];
      if (Anew[i] == width) {
	if (bCond == PERIODIC_BOUNDARY)
	  Anew[i] = 0;    // periodic boundary condition
	else
	  return;  // no flux boundary, cannot more right
      }
    }
  } else
    return;

  // check if move is valid, if so then do it
  if (check_overlap(grid, width, height, Anew, Bnew, spatialExtent[0], spatialExtent[1], cellSize)) {
    if ([cellColony validateMoveForCell: self newXExtent: Anew newYExtent: Bnew cellSize: cellSize])
      rewrite_cell(grid, width, height, Anew, Bnew, spatialExtent[0], spatialExtent[1], cellSize, self);
  }
}

- (void)roundUp
{
  unsigned int *newExtent[2];
  NSDictionary *d = [cellColony spatialGrid];
  NSNumber *n = [d objectForKey: @"width"];
  int width = [n intValue];
  n = [d objectForKey: @"height"];
  int height = [n intValue];
  Grid2DArray *m = [d objectForKey: @"matrix"];
  id (*grid)[height][width] = [m getMatrix];

  //printf("roundUp %d\n", self);


  newExtent[0] = (unsigned int *)malloc(sizeof(unsigned int) * 5);
  newExtent[1] = (unsigned int *)malloc(sizeof(unsigned int) * 5);

  // keep nucleus
  int i;
  for (i = 0; i < 4; ++i) {
    newExtent[0][i] = spatialExtent[0][i];
    newExtent[1][i] = spatialExtent[1][i];
  }

  //printf("%x\n", self);
  /*
  for (i = 0; i < 7; i++) {
    printf("%d %d\n", newExtent[1][i], newExtent[0][i]);
    printf("%x\n", (*grid)[newExtent[1][i]][newExtent[0][i]]);
    }
  */

  // clear protusion areas
  for (i = 4; i < 7; i++) {
    //printf("%d %d\n", spatialExtent[1][i], spatialExtent[0][i]);
    //printf("%x\n", (*grid)[spatialExtent[1][i]][spatialExtent[0][i]]);
    (*grid)[spatialExtent[1][i]][spatialExtent[0][i]] = nil;
    }

  /*
  for (i = 0; i < 4; i++) {
    printf("%d %d\n", newExtent[1][i], newExtent[0][i]);
    printf("%x\n", (*grid)[newExtent[1][i]][newExtent[0][i]]);
    printf("%d %d\n", spatialExtent[1][i], spatialExtent[0][i]);
    printf("%x\n", (*grid)[spatialExtent[1][i]][spatialExtent[0][i]]);
    }
  */

  // keep one of three protusions
  int r = (int)(RAND_NUM * 3.0);
  switch (r) {
  case 0:
    newExtent[0][4] = spatialExtent[0][4];
    newExtent[1][4] = spatialExtent[1][4];
    (*grid)[spatialExtent[1][4]][spatialExtent[0][4]] = self;
    (*grid)[spatialExtent[1][5]][spatialExtent[0][5]] = nil;
    (*grid)[spatialExtent[1][6]][spatialExtent[0][6]] = nil;
    break;
  case 1:
    newExtent[0][4] = spatialExtent[0][5];
    newExtent[1][4] = spatialExtent[1][5];
    (*grid)[spatialExtent[1][4]][spatialExtent[0][4]] = nil;
    (*grid)[spatialExtent[1][5]][spatialExtent[0][5]] = self;
    (*grid)[spatialExtent[1][6]][spatialExtent[0][6]] = nil;
    break;
  case 2:
    newExtent[0][4] = spatialExtent[0][6];
    newExtent[1][4] = spatialExtent[1][6];
    //printf("%x\n", (*grid)[spatialExtent[1][5]][spatialExtent[0][5]]);
    //printf("%x\n", (*grid)[spatialExtent[1][6]][spatialExtent[0][6]]);
    (*grid)[spatialExtent[1][4]][spatialExtent[0][4]] = nil;
    (*grid)[spatialExtent[1][5]][spatialExtent[0][5]] = nil;
    (*grid)[spatialExtent[1][6]][spatialExtent[0][6]] = self;
    break;
  }

  cellSize = 5;
  free(spatialExtent[0]);
  free(spatialExtent[1]);
  spatialExtent[0] = newExtent[0];
  spatialExtent[1] = newExtent[1];

  /*
  for (i = 0; i < cellSize; i++) {
    printf("%d %d\n", spatialExtent[1][i], spatialExtent[0][i]);
    printf("%x\n", (*grid)[spatialExtent[1][i]][spatialExtent[0][i]]);
    }
  */
}

@end
