/*
 CellColony.m
 
 Colony of biological cells.
 
 Copyright (C) 2005 Scott Christley
 
 Author: Scott Christley <schristl@nd.edu>
 Date: 2005
 
 This file is part of LimbChondrogenesis.
 */

#import "CellColony.h"
#import "BiologicalCell.h"
#import "ReactionDiffusion.h"
#import "Grid2DArray.h"

#include "internal.h"

@implementation CellColony

- initWithSpatialGrid: (NSDictionary *)aGrid
{
	[super init];
	
	spatialGrid = aGrid;
	[spatialGrid retain];
	actMemory = NO;
	
	return self;
}

- (NSArray *)createColonyWithCellNumber: (unsigned int)count
{
	return [self createColonyWithCellNumber: count ofClass: [BiologicalCell class]];
}

- (NSArray *)createColonyWithCellNumber: (unsigned int)count ofClass: (Class)cellClass
{
	cells = [NSMutableArray new];
	
	NSNumber *n = [spatialGrid objectForKey: @"height"];
	int h = [n intValue];
	n = [spatialGrid objectForKey: @"width"];
	int w = [n intValue];
	int hBound, wBound;
	
	// 3x3 area for cell
	int hRange = (h - 1) / 3;
	int wRange = (w - 1) / 3;
	if ((hRange * wRange) > count) {
		// plenty of space
		hRange = count / wRange;
		hBound = (h - 1) / hRange;
		wBound = (w - 1) / wRange;
	} else {
		// not enough space to fit all cells
		if ((hRange * wRange) != count)
			printf("Can only fit %d cells!\n", hRange*wRange);
		hBound = 3;
		wBound = 3;
	}
	//int range = (int) sqrt(count);
	//int hBound = h / range;
	//int wBound = w / range;
	printf("hRange=%d, wRange=%d, hBound=%d, wBound=%d\n", hRange, wRange, hBound, wBound);
	
	// create the cells
	int i, j, x, y;
	for (i = 0; i < hRange; i++)
		for (j = 0; j < wRange; j++) {
			y = 1 + hBound * i;
			x = 1 + wBound * j;
			//printf("cell %d %d\n", x, y);
			id cell = [[cellClass alloc] initWithColony: self ofSize: 7 atX: x atY: y ofType: 0];
			[cells addObject: cell];
		}
			
	return cells;
}

- (NSDictionary *)spatialGrid
{
	return spatialGrid;
}

- (void)recordActivatorMemory: (BOOL)aFlag;
{
	actMemory = aFlag;
}

//
//
//
- (NSArray *)randomizedCellList
{
	return cells;
}

- (void)cellsPerform: (BOOL)randomize withInvocation: (NSInvocation *)anInvocation
{
	int i;
	NSArray *c;

	if (randomize)
		c = [self randomizedCellList];
	else
		c = cells;

	for (i = 0; i < [c count]; ++i) {
		id cell = [c objectAtIndex: i];
		[anInvocation invokeWithTarget: cell];
	}
}

//
// cell diffusion
//

- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb forCells: (NSArray *)cellList
{
	int i, j;
	
	int cnt = [cellList count];
	for (i = 0; i < aCnt; i++) {
		for (j = 0; j < cnt; j++) {
			// randomly pick a cell
			double r = RAND_NUM;
			double z = r * cnt;
			int t = floor(z);
			BiologicalCell *cell = [cellList objectAtIndex: t];
			
			// change shape
			if (sProb > 0.0) {
				r = RAND_NUM;
				if (r < sProb) [cell changeShape];
			}
			
			// move
			[cell moveWithProbability: mProb];
		}
	}
}

- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb
{
	[self performDiffusion: aCnt withShapeProbability: sProb withMoveProbability: mProb forCells: cells];
}

- (void)performDiffusion: (unsigned int)aCnt withMoveProbability: (double)mProb
{
	[self performDiffusion: aCnt withShapeProbability: 1.0 withMoveProbability: mProb forCells: cells];
}

- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb forCellType: (int)cellType
{
	//NSMutableArray *types = [[NSMutableArray new] autorelease];
	NSMutableArray *types = [NSMutableArray new];
	int i;
	
	// collect the cells of the right type
	int cnt = [cells count];
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		if ([cell cellType] == cellType)
			[types addObject: cell];
	}
	
	cnt = [types count];
	if (cnt == 0) return;
	
	[self performDiffusion: aCnt withShapeProbability: sProb withMoveProbability: mProb forCells: types];
}

- (void)performDiffusion: (unsigned int)aCnt withShapeProbability: (double)sProb
     withMoveProbability: (double)mProb forCellInvocation: (NSInvocation *)anInvocation
{
	//NSMutableArray *types = [[NSMutableArray new] autorelease];
	NSMutableArray *types = [NSMutableArray new];
	int i;
	
	// collect the cells of the right type
	int cnt = [cells count];
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		BOOL flag;
		[anInvocation invokeWithTarget: cell];
		[anInvocation getReturnValue: &flag];
		if (flag)
			[types addObject: cell];
	}
	
	cnt = [types count];
	if (cnt == 0) return;
	
	[self performDiffusion: aCnt withShapeProbability: sProb withMoveProbability: mProb forCells: types];
}

- (void)checkStuckOnFibronectin: (NSDictionary *)aGrid
{
	int i, j;
	
	// spatial grid for fibronectin
	NSNumber *n = [aGrid objectForKey: @"width"];
	int fibroWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int fibroHeight = [n intValue];
	n = [aGrid objectForKey: @"scale"];
	int scale = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	id (*fibroGrid)[fibroHeight][fibroWidth] = [m getMatrix];
	
	// If the majority of cell is spatially on top
	// of fibronectin then consider it stuck
	int cnt = [cells count];
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		if ([cell stuckOnFibronectin]) continue;
		
		int cellSize = [cell cellSize];
		unsigned int *xExtent = [cell xSpatialExtent];
		unsigned int *yExtent = [cell ySpatialExtent];
		
		int fibroCnt = 0;
		for (j = 0; j < cellSize; ++j) {
			// scale coordinates to fibronectin grid
			int h = (yExtent[j] * fibroHeight) / height;
			int w = (xExtent[j] * fibroWidth) / width;
			int k, l;
			for (k = 0; k < scale; ++k)
				for (l = 0; l < scale; ++l) {
					if ((*fibroGrid)[h + k][w + l] > 0)
						++fibroCnt;
				}
		}
			
			int test = (cellSize * scale * scale) / 2;
			if (fibroCnt > test) {
				[cell setStuckOnFibronectin: YES];
				//printf("Cell %d %d stuck test: %d fibroCnt: %d\n",
				//     xExtent[0], yExtent[0], test, fibroCnt);
			}
	}
}

//
// delegate methods for cell diffusion
//

- (void)setFibronectinGrid: (NSDictionary *)aGrid
{
	fibronectinGrid = aGrid;
	
	// cache values
	if (fibronectinGrid) {
		NSNumber *n = [fibronectinGrid objectForKey: @"width"];
		fibronectinWidth = [n intValue];
		n = [fibronectinGrid objectForKey: @"height"];
		fibronectinHeight = [n intValue];
		n = [fibronectinGrid objectForKey: @"scale"];
		fibronectinScale = [n intValue];
	}
}

- (BOOL)validateMoveForCell: (id)aCell newXExtent: (unsigned int *)xExtent newYExtent: (unsigned int *)yExtent
				   cellSize: (int)cellSize;
{
	// only concerned about cells stuck on fibronectin
	if (![aCell stuckOnFibronectin])
		return YES;
	
	Grid2DArray *m = [fibronectinGrid objectForKey: @"matrix"];
	id (*fibroGrid)[fibronectinHeight][fibronectinWidth] = [m getMatrix];
	
	// If the majority of cell is spatially on top
	// of fibronectin then move is okay, still
	// considered stuck on fibronectin, otherwise
	// disallow move.
	int j, fibroCnt = 0;
	for (j = 0; j < cellSize; ++j) {
		// scale coordinates to fibronectin grid
		int h = (yExtent[j] * fibronectinHeight) / height;
		int w = (xExtent[j] * fibronectinWidth) / width;
		int k, l;
		for (k = 0; k < fibronectinScale; ++k)
			for (l = 0; l < fibronectinScale; ++l) {
				if ((*fibroGrid)[h + k][w + l] > 0)
					++fibroCnt;
			}
	}
		
		int test = (cellSize * fibronectinScale * fibronectinScale) / 2;
		if (fibroCnt > test) {
			// still stuck on fibronectin
			return YES;
		} else {
			// tried to move off!
			// check probability of it moving off
			//printf("Cell %d %d tried to move off fibronectin! stuck test: %d fibroCnt: %d\n",
			//   xExtent[0], yExtent[0], test, fibroCnt);
			NSString *s = [simParameters objectForKey: @"fibronectinCellProb"];
			double fibronectinCellProb = [s doubleValue];
			if (RAND_NUM < fibronectinCellProb)
				return NO;
			else
				return YES;
		}
}

//
// cell differentiation
//

- (BOOL)checkThreshold: (int)aValue forGrid: (NSDictionary *)aGrid
{
	int i, j, cnt = [cells count];
	BOOL found = NO;
	
	NSNumber *n = [aGrid objectForKey: @"width"];
	int checkWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int checkHeight = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	int (*checkGrid)[checkHeight][checkWidth] = [m getMatrix];
	
	m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	// loop through the cells and initialize
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		[cell setActivatorSum: 0];
	}
	
	// record activator levels
	for (i = 0; i < checkHeight; ++i) {
		for (j = 0; j < checkWidth; ++j) {
			// scale coordinates to cell grid
			int h = (i * height) / checkHeight;
			int w = (j * width) / checkWidth;
			
			// differentiate cell if above threshold
			if ((*cellGrid)[h][w] != nil) {
				//if ((*checkGrid)[i][j] >= aValue) {
				BiologicalCell *cell = (*cellGrid)[h][w];
				[cell setActivatorSum: [cell activatorSum] + (*checkGrid)[i][j]];
				//[cell setCellType: 1];
				//found = YES;
				}
			}
		}
	
	// loop through the cells and check the threshold
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		//if ([cell activatorSum] != 0) printf("activatorSum: %d\n", [cell activatorSum]);
		if ([cell activatorSum] >= aValue) {
			[cell setCellType: 1];
			found = YES;
		}
	}
	
	return found;
	}

- (BOOL)checkDoubleThreshold: (int)aValue forGrid: (NSDictionary *)aGrid
{
	int i, j, cnt = [cells count];
	BOOL found = NO;
	
	NSNumber *n = [aGrid objectForKey: @"width"];
	int checkWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int checkHeight = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	double (*checkGrid)[checkHeight][checkWidth] = [m getMatrix];
	
	m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	// loop through the cells and initialize
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		[cell setActivatorSum: 0];
	}
	
	// record activator levels
	for (i = 0; i < checkHeight; ++i) {
		for (j = 0; j < checkWidth; ++j) {
			// scale coordinates to cell grid
			int h = (i * height) / checkHeight;
			int w = (j * width) / checkWidth;
			
			// differentiate cell if above threshold
			if ((*cellGrid)[h][w] != nil) {
				//if ((*checkGrid)[i][j] >= aValue) {
				BiologicalCell *cell = (*cellGrid)[h][w];
				[cell setActivatorSum: [cell activatorSum] + (int)(*checkGrid)[i][j]];
				//[cell setCellType: 1];
				//found = YES;
				}
			}
		}
	
	// loop through the cells and check the threshold
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		//if ([cell activatorSum] != 0) printf("activatorSum: %d\n", [cell activatorSum]);
		if ([cell activatorSum] >= aValue) {
			[cell setCellType: 1];
			found = YES;
		}
	}
	
	return found;
	}

- (BOOL)checkThreshold: (int)aValue withInvocation: (NSInvocation *)anInvocation
{
	int i, cnt = [cells count];
	BOOL found = NO;
	
	// loop through the cells and check the threshold
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		int value;
		[anInvocation invokeWithTarget: cell];
		[anInvocation getReturnValue: &value];
		if (value > aValue) {
			[cell setCellType: 1];
			found = YES;
		}
	}
	
	return found;
}

- (BOOL)checkThresholdForGrid: (NSDictionary *)aGrid
{
	int i, j, cnt = [cells count];
	BOOL found = NO;
	
	NSNumber *n = [aGrid objectForKey: @"width"];
	int checkWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int checkHeight = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	int (*checkGrid)[checkHeight][checkWidth] = [m getMatrix];
	
	m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	// loop through the cells and initialize
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		[cell setActivatorSum: 0];
	}
	
	// record activator levels
	for (i = 0; i < checkHeight; ++i) {
		for (j = 0; j < checkWidth; ++j) {
			// scale coordinates to cell grid
			int h = (i * height) / checkHeight;
			int w = (j * width) / checkWidth;
			
			// differentiate cell if above threshold
			if ((*cellGrid)[h][w] != nil) {
				//if ((*checkGrid)[i][j] >= aValue) {
				BiologicalCell *cell = (*cellGrid)[h][w];
				[cell setActivatorSum: [cell activatorSum] + (*checkGrid)[i][j]];
				//[cell setCellType: 1];
				//found = YES;
				}
			}
		}
	
	// loop through the cells and check the threshold
	for (i = 0; i < cnt; ++i) {
		BiologicalCell *cell = [cells objectAtIndex: i];
		//if ([cell activatorSum] != 0) printf("activatorSum: %d\n", [cell activatorSum]);
    //printf("activatorThreshold: %d\n", [cell activatorThreshold]);
		if ([cell activatorSum] >= [cell activatorThreshold]) {
			[cell setCellType: 1];
			found = YES;
		}
	}
	
	return found;
}

- (void)cellRoundUp: (int)aValue forGrid: (NSDictionary *)aGrid
{
	int i, j;
	
	NSNumber *n = [aGrid objectForKey: @"width"];
	int checkWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int checkHeight = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	int (*checkGrid)[checkHeight][checkWidth] = [m getMatrix];
	
	m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	for (i = 0; i < checkHeight; ++i) {
		for (j = 0; j < checkWidth; ++j) {
			// scale coordinates to cell grid
			int h = (i * height) / checkHeight;
			int w = (j * width) / checkWidth;
			
			// round up cell if above threshold
			if ((*cellGrid)[h][w] != nil)
				if ((*checkGrid)[i][j] >= aValue) {
					BiologicalCell *cell = (*cellGrid)[h][w];
					if ([cell cellSize] != 5)
						[cell roundUp];
				}
		}
	}
}

- (void)cellDoubleRoundUp: (int)aValue forGrid: (NSDictionary *)aGrid
{
	int i, j;
	
	NSNumber *n = [aGrid objectForKey: @"width"];
	int checkWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int checkHeight = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	double (*checkGrid)[checkHeight][checkWidth] = [m getMatrix];
	
	m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	for (i = 0; i < checkHeight; ++i) {
		for (j = 0; j < checkWidth; ++j) {
			// scale coordinates to cell grid
			int h = (i * height) / checkHeight;
			int w = (j * width) / checkWidth;
			
			// round up cell if above threshold
			if ((*cellGrid)[h][w] != nil)
				if ((*checkGrid)[i][j] >= aValue) {
					BiologicalCell *cell = (*cellGrid)[h][w];
					if ([cell cellSize] != 5)
						[cell roundUp];
				}
		}
	}
}

- (void)productionForGrid: (NSDictionary *)aGrid
{
	int i, j;
	
	// production grid
	NSNumber *n = [aGrid objectForKey: @"width"];
	int prodWidth = [n intValue];
	n = [aGrid objectForKey: @"height"];
	int prodHeight = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	int (*prodGrid)[prodHeight][prodWidth] = [m getMatrix];
	n = [aGrid objectForKey: @"boundaryCondition"];
	int bCond = [n intValue];
	
	// cell grid
	m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	// reaction parameters
	NSString *s = [simParameters objectForKey: @"fibronectinProduceProb"];
	double productionProb = [s doubleValue];
	s = [simParameters objectForKey: @"fibronectinDiffusion"];
	double initialDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"maxFibronectin"];
	int maxProduce = [s intValue];
	s = [simParameters objectForKey: @"fibronectinDelay"];
	int fibroDelay = [s intValue];
	
	for (i = 0; i < prodHeight; ++i) {
		for (j = 0; j < prodWidth; ++j) {
			// scale coordinates to cell grid
			int h = (i * height) / prodHeight;
			int w = (j * width) / prodWidth;
			
			// production only at cell
			if ((*cellGrid)[h][w] != nil) {
				BiologicalCell *cell = (*cellGrid)[h][w];
				if ([cell cellType] == 1) {
					
					// delay production
					[cell incrementFibroProduction: 1];
					if ([cell fibroProduction] < fibroDelay) continue;
					
					double r = RAND_NUM;
					if (r < productionProb) {
						int y = i;
						int x = j;
						
						// initial diffusion
						int r1 = (int) (RAND_NUM * (4.0 / initialDiffusion));
						switch(r1) {
							case 0:
								// diffuse right
								++x;
								if (x == prodWidth) {
									if (bCond == PERIODIC_BOUNDARY) x = 0;
									else x = prodWidth - 1;
								}
									break;
							case 1:
								// diffuse up
								--y;
								if (y == -1) {
									if (bCond == PERIODIC_BOUNDARY) y = prodHeight - 1;
									else y = 0;
								}
									break;
							case 2:
								// diffuse left
								--x;
								if (x == -1) {
									if (bCond == PERIODIC_BOUNDARY) x = prodWidth - 1;
									else x = 0;
								}
									break;
							case 3:
								// diffuse down
								++y;
								if (y == prodHeight) {
									if (bCond == PERIODIC_BOUNDARY) y = 0;
									else y = prodHeight - 1;
								}
									break;
						}
						
						// produce
						(*prodGrid)[y][x] += 1;
						if ((*prodGrid)[y][x] > maxProduce)
							(*prodGrid)[y][x] = maxProduce;
					}
				}
			}
		}
	}
	
}

//
// delegate methods for reaction-diffusion
//

- (void)setupReaction: (MultiScale2DGrid *)aGrid withActivatorGrid: (NSDictionary *)actGrid
	withInhibitorGrid: (NSDictionary *)inhibGrid
{
	grid = aGrid;
	activatorGrid = actGrid;
	inhibitorGrid = inhibGrid;
	
	NSNumber *n = [activatorGrid objectForKey: @"height"];
	moleculeHeight = [n intValue];
	n = [activatorGrid objectForKey: @"width"];
	moleculeWidth = [n intValue];
	
	n = [spatialGrid objectForKey: @"height"];
	height = [n intValue];
	n = [spatialGrid objectForKey: @"width"];
	width = [n intValue];
	
	// reaction parameters
	NSString *s = [simParameters objectForKey: @"cellActivator"];
	cellActivator = [s intValue];
}

- (BOOL)productionOfActivator: (int *)actProduce ofInhibitor: (int *)inhibProduce atX: (int)x atY: (int)y
{
	Grid2DArray *m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	// scale coordinates to cell grid
	int w = (x * width) / moleculeWidth;
	int h = (y * height) / moleculeHeight;
	//printf("%d %d\n", w, h);
	
	// production occurs only where there is a cell
	if ((*cellGrid)[h][w] != nil) {
		// cell produces activator
		*actProduce += cellActivator;
		//printf("cell (%d %d) -> (%d %d): %d %d\n", y, x, h, w, *actProduce, *inhibProduce);
	} else {
		*actProduce = 0;
		*inhibProduce = 0;
	}
	
	// record activator production
	if (actMemory) {
		BiologicalCell *cell = (*cellGrid)[h][w];
		[cell recordActivatorProduce: *actProduce];
	}
	
	return YES;
}

- (BOOL)doubleProductionOfActivator: (double *)actProduce ofInhibitor: (double *)inhibProduce atX: (int)x atY: (int)y
{
	Grid2DArray *m = [spatialGrid objectForKey: @"matrix"];
	id (*cellGrid)[height][width] = [m getMatrix];
	
	// scale coordinates to cell grid
	int w = (x * width) / moleculeWidth;
	int h = (y * height) / moleculeHeight;
	//printf("%d %d\n", w, h);
	
	// production occurs only where there is a cell
	if ((*cellGrid)[h][w] != nil) {
		// cell produces activator
		*actProduce += cellActivator;
		//printf("cell (%d %d) -> (%d %d): %d %d\n", y, x, h, w, *actProduce, *inhibProduce);
	} else {
		*actProduce = 0;
		*inhibProduce = 0;
	}
	
	// record activator production
	//  if (actMemory) {
	//    BiologicalCell *cell = (*cellGrid)[h][w];
	//    [cell recordActivatorProduce: *actProduce];
	//  }
	
	return YES;
}

@end
