/*
  MultiScale2DGrid.m

  Implementation of multiple 2D spatial grid overlays at different scales

  Copyright (C) 2005 Scott Christley

  Author: Scott Christley <schristl@nd.edu>
  Date: 2005
   
  This file is part of LimbChondrogenesis.
*/

#import "MultiScale2DGrid.h"
#import "Grid2DArray.h"

#include "internal.h"

@implementation MultiScale2DGrid

- initWithWidth: (unsigned int)aWidth andHeight: (unsigned int)aHeight
{
  [super init];

  width = aWidth;
  height = aHeight;
  grids = [[NSMutableDictionary new] retain];

  return self;
}

- (void)dealloc
{
  [grids release];
  [super dealloc];
}

- (NSDictionary *)newGridWithName: (NSString *)name atScale: (unsigned int)scale
		       withEncode: (NSString *)encode
{
  NSMutableDictionary *newGrid = [NSMutableDictionary new];

  int w = width * scale;
  int h = height * scale;

  [newGrid setObject: [[NSNumber alloc] initWithInt: w] forKey: @"width"];
  [newGrid setObject: [[NSNumber alloc] initWithInt: h] forKey: @"height"];
  [newGrid setObject: [[NSNumber alloc] initWithInt: scale] forKey: @"scale"];
  [newGrid setObject: encode forKey: @"encode"];
  id matrix = [[Grid2DArray alloc] initWithWidth: w andHeight: h andEncode: encode];
  [matrix getMatrix];
  [newGrid setObject: matrix forKey: @"matrix"];
  [grids setObject: newGrid forKey: name];

  return newGrid;
}

- (void)removeGridWithName: (NSString *)name
{
}

- (NSDictionary *)gridWithName: (NSString *)name
{
  return [grids objectForKey: name];
}

- (int)maxValueForIntGrid: (NSString *)gridName
{
  NSDictionary *aGrid = [self gridWithName: gridName];

  if (!aGrid) return 0;

  NSNumber *n = [aGrid objectForKey: @"height"];
  int h = [n intValue];
  n = [aGrid objectForKey: @"width"];
  int w = [n intValue];
  Grid2DArray *m = [aGrid objectForKey: @"matrix"];
  int (*grid)[h][w] = [m getMatrix];

  int i, j;
  int max = INT_MIN;
  for (i = 0; i < h; ++i)
    for (j = 0; j < w; ++j)
      if ((*grid)[i][j] > max)
	max = (*grid)[i][j];

  return max;
}

- (double)maxValueForDoubleGrid: (NSString *)gridName
{
  NSDictionary *aGrid = [self gridWithName: gridName];

  if (!aGrid) return 0;

  NSNumber *n = [aGrid objectForKey: @"height"];
  int h = [n intValue];
  n = [aGrid objectForKey: @"width"];
  int w = [n intValue];
  Grid2DArray *m = [aGrid objectForKey: @"matrix"];
  double (*grid)[h][w] = [m getMatrix];

  int i, j;
  double max = -1;
  for (i = 0; i < h; ++i)
    for (j = 0; j < w; ++j)
      if ((*grid)[i][j] > max)
	max = (*grid)[i][j];

  return max;
}

- (void)dumpGrid: (NSString *)gridName toFile: (FILE *)aFile
{
  [self dumpGrid: gridName toFile: aFile withInvocation: nil];
}

- (void)dumpGrid: (NSString *)gridName toFile: (FILE *)aFile withInvocation: (NSInvocation *)anInvocation
{
  NSDictionary *aGrid = [grids objectForKey: gridName];
  NSNumber *n = [aGrid objectForKey: @"height"];
  int h = [n intValue];
  n = [aGrid objectForKey: @"width"];
  int w = [n intValue];
  Grid2DArray *m = [aGrid objectForKey: @"matrix"];
  NSString *encode = [aGrid objectForKey: @"encode"];

  int zero = 0;
  int one = 1;
  int i,j;
  if ([encode isEqual: idEncode]) {
    id (*grid)[h][w] = [m getMatrix];
    for (i = 0;i < h; ++i)
      for (j = 0;j < w; ++j)
	if ((*grid)[i][j] != nil) {
	  if (anInvocation) {
	    int value;
	    [anInvocation invokeWithTarget: (*grid)[i][j]];
	    [anInvocation getReturnValue: &value];
	    //printf("value=%d\n", value);
	    fwrite(&value, sizeof(int), 1, aFile);
	  } else {
	    if ([(*grid)[i][j] respondsToSelector: @selector(intValue)]) {
	      int value = [(*grid)[i][j] intValue];
	      fwrite(&value, sizeof(int), 1, aFile);
	    } else
	      fwrite(&one, sizeof(int), 1, aFile);
	  }
	} else
	  fwrite(&zero, sizeof(int), 1, aFile);
  } else if ([encode isEqual: intEncode]) {
    int (*grid)[h][w] = [m getMatrix];
    for (i = 0;i < h; ++i)
      for (j = 0;j < w; ++j)
	fwrite(&((*grid)[i][j]), sizeof(int), 1, aFile);
  } else if ([encode isEqual: longEncode]) {
    long (*grid)[h][w] = [m getMatrix];
    for (i = 0;i < h; ++i)
      for (j = 0;j < w; ++j)
	fwrite(&((*grid)[i][j]), sizeof(long), 1, aFile);
  } else if ([encode isEqual: doubleEncode]) {
    double (*grid)[h][w] = [m getMatrix];
    for (i = 0;i < h; ++i)
      for (j = 0;j < w; ++j)
	fwrite(&((*grid)[i][j]), sizeof(double), 1, aFile);
  } else if ([encode isEqual: boolEncode]) {
    BOOL (*grid)[h][w] = [m getMatrix];
    for (i = 0;i < h; ++i)
      for (j = 0;j < w; ++j)
	if ((*grid)[i][j] == YES)
	  fwrite(&one, sizeof(int), 1, aFile);
	else
	  fwrite(&zero, sizeof(int), 1, aFile);
  } else {
    // throw exception or something
  }
}

@end
