/*
 ReactionDiffusion.m
 
 reaction-diffusion chemical process.
 
 Copyright (C) 2005 Scott Christley
 
 Author: Scott Christley <schristl@nd.edu>
 Date: 2005
 
 This file is part of LimbChondrogenesis.
 */

#include <Foundation/Foundation.h>
#include "MultiScale2DGrid.h"
#include "Grid2DArray.h"
#include "ReactionDiffusion.h"

#include "internal.h"

@implementation ReactionDiffusion

- initForGrid: (MultiScale2DGrid *)aGrid withActivatorGrid: (NSString *)aName
withInhibitorGrid: (NSString *)iName
{
	[super init];
	
	grid = aGrid;
	activatorName = aName;
	inhibitorName = iName;
	flagName = nil;
	delegate = nil;
	delegatePreReaction = NO;
	delegateProduction = NO;
	delegateAfterCalculate = NO;
	delegatePostReaction = NO;
	
	return self;
}

- (void)setDelegate: (id)anObj
{
	[delegate release];
	delegate = anObj;
	[delegate retain];
	
	// cache methods delegate responds to
	if ([delegate respondsToSelector: @selector(preReaction)])
		delegatePreReaction = YES;
	else
		delegatePreReaction = NO;
	
	if ([delegate respondsToSelector: @selector(productionOfActivator:ofInhibitor:atX:atY:)])
		delegateProduction = YES;
	else
		delegateProduction = NO;
	
	if ([delegate respondsToSelector: @selector(afterCalculate)])
		delegateAfterCalculate = YES;
	else
		delegateAfterCalculate = NO;
	
	if ([delegate respondsToSelector: @selector(postReaction)])
		delegatePostReaction = YES;
	else
		delegatePostReaction = NO;
}

- (void)performReaction
{
	// get activator/inhibitor grids
	NSDictionary *aGrid = [grid gridWithName: activatorName];
	NSDictionary *iGrid = [grid gridWithName: inhibitorName];
	NSNumber *n = [aGrid objectForKey: @"height"];
	int height = [n intValue];
	n = [aGrid objectForKey: @"width"];
	int width = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	int (*activatorGrid)[height][width] = [m getMatrix];
	m = [iGrid objectForKey: @"matrix"];
	int (*inhibitorGrid)[height][width] = [m getMatrix];
	
	// reaction parameters
	NSString *s = [simParameters objectForKey: @"activatorOnActivator"];
	double activatorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"activatorOnInhibitor"];
	double activatorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnActivator"];
	double inhibitorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnInhibitor"];
	double inhibitorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"maxActivatorProduce"];
	int maxActivatorProduce = [s intValue];
	s = [simParameters objectForKey: @"maxInhibitorProduce"];
	int maxInhibitorProduce = [s intValue];
	//printf("activator: %lf %lf %d\n", activatorOnActivator, activatorOnInhibitor, maxActivatorProduce);
	//printf("inhibitor: %lf %lf %d\n", inhibitorOnActivator, inhibitorOnInhibitor, maxInhibitorProduce);
	
	// setup delegate
	[delegate setupReaction: grid withActivatorGrid: aGrid withInhibitorGrid: iGrid];
	
	// perform reaction
	int i, j;
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			
			if (delegatePreReaction)
				if (![delegate preReaction])
					continue;
			
			// activator produce activator
			double activator = (double)(*activatorGrid)[i][j];
			int actProduce = (int)(activator * activatorOnActivator + 0.5);
			if ((maxActivatorProduce != -1) && (actProduce > maxActivatorProduce))
				actProduce = maxActivatorProduce;
			
			// activator produce inhibitor
			double inhibitor = (double)(*inhibitorGrid)[i][j];
			int inhibProduce = (int)(activator * activatorOnInhibitor + 0.5);
			if ((maxInhibitorProduce != -1) && (inhibProduce > maxInhibitorProduce))
				inhibProduce = maxInhibitorProduce;
			
			if (delegateProduction)
				if (![delegate productionOfActivator: &actProduce ofInhibitor: &inhibProduce atX: j atY: i])
					continue;
			
			// inhibitor cancel activator
			double actCancel = inhibitor * inhibitorOnActivator;
			if ((actCancel + activator) < 0)
				actCancel = -activator;
			int cancelActivator = (int)((-actCancel) + 0.5);
			
			// inhibitor decays
			int inhibDecay = (int)(actCancel * inhibitorOnInhibitor + 0.5);
			
			if (delegateAfterCalculate)
				[delegate afterCalculate];
			
			// update activator/inhibitor for this pixel
			(*activatorGrid)[i][j] = (*activatorGrid)[i][j] + actProduce - cancelActivator;
			(*inhibitorGrid)[i][j] = (*inhibitorGrid)[i][j] + inhibProduce - inhibDecay;
			
			/*
			 if ((*activatorGrid)[i][j] > 0) {
				 printf("%d %d: %lf + %d - %d = %d\n", i, j, activator, actProduce, cancelActivator, (*activatorGrid)[i][j]);
				 printf("%d %d: %lf + %d - %d = %d\n", i, j, inhibitor, inhibProduce, inhibDecay, (*inhibitorGrid)[i][j]);
			 } */
			
			if (delegatePostReaction)
				[delegate postReaction];
		}
	}
	
}

- (void)performDiffusion
{
	// get activator/inhibitor grids
	NSDictionary *aGrid = [grid gridWithName: activatorName];
	NSDictionary *iGrid = [grid gridWithName: inhibitorName];
	NSNumber *n = [aGrid objectForKey: @"height"];
	int height = [n intValue];
	n = [aGrid objectForKey: @"width"];
	int width = [n intValue];
	
	// diffusion parameters
	NSString *s = [simParameters objectForKey: @"activatorDiffusion"];
	double activatorDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionOnFibronectin"];
	double activatorDiffusionOnFibronectin = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionTime"];
	int activatorDiffusionTime = [s intValue];
	s = [simParameters objectForKey: @"inhibitorDiffusion"];
	double inhibitorDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionTime"];
	int inhibitorDiffusionTime = [s intValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionOnFibronectin"];
	double inhibitorDiffusionOnFibronectin = [s doubleValue];
	BOOL use_alt = NO;
	//printf("%lf %lf %d %lf %d\n", activatorDiffusion, activatorDiffusionOnFibronectin,
	// activatorDiffusionTime, inhibitorDiffusion, inhibitorDiffusionTime);
	
	int i, j, k, RND;
	int numTot, numRight, numLeft, numUp, numDown;
	double num, useDiffAll, useDiffOne;
	double diffRateAll, diffRateOne;
	double altDiffRateAll, altDiffRateOne;
	int matrix[height][width];
	int molecule;
	int diffusionTime;
	int bCond;
	
	//printf("performDiffusion\n");
	
	// diffusion for both activator and inhibitor
	for (molecule = 0; molecule < 2; ++molecule) {
		int (*theGrid)[height][width];
		
		if (molecule == 0) {
			// activator
			Grid2DArray *m = [aGrid objectForKey: @"matrix"];
			theGrid = [m getMatrix];
			diffusionTime = activatorDiffusionTime;
			
			// equal rate of diffusion for each direction
			diffRateAll = activatorDiffusion;
			diffRateOne = activatorDiffusion / 4.0;
			// alternate diffusion rate on fibronectin
			altDiffRateAll = activatorDiffusionOnFibronectin;
			altDiffRateOne = activatorDiffusionOnFibronectin / 4.0;
			// boundary condition
			n = [aGrid objectForKey: @"boundaryCondition"];
			bCond = [n intValue];
		} else {
			// inhibitor
			Grid2DArray *m = [iGrid objectForKey: @"matrix"];
			theGrid = [m getMatrix];
			diffusionTime = inhibitorDiffusionTime;
			
			// equal rate of diffusion for each direction
			diffRateAll = inhibitorDiffusion;
			diffRateOne = inhibitorDiffusion / 4.0;
			// alternate diffusion rate on fibronectin
			altDiffRateAll = inhibitorDiffusionOnFibronectin;
			altDiffRateOne = inhibitorDiffusionOnFibronectin / 4.0;
			// boundary condition
			n = [iGrid objectForKey: @"boundaryCondition"];
			bCond = [n intValue];
		}
		
		// perform diffusion some number of times
		for (k = 0; k < diffusionTime; ++k) {
			
			// Make copy of current grid
			for (i = 0; i < height; i++)
				for (j = 0; j < width; j++)
					matrix[i][j] = (*theGrid)[i][j];
			
			// diffuse
			for (i = 0; i < height; i++) {
				for (j = 0; j < width; j++) {
					
					// equal rate of diffusion for each direction
					useDiffAll = diffRateAll;
					useDiffOne = diffRateOne;
					// alternate diffusion rate on fibronectin
					if (use_alt) {
						//if (FN[i][j] > 0) {
						useDiffAll = altDiffRateAll;
						useDiffOne = altDiffRateOne;
						//}
					}
					
					if (matrix[i][j] > 0) {
						// determine amount
						num = (double)matrix[i][j];
						numTot = (int) floor((num * useDiffAll) + 0.5);
						//printf("%f %d\n", num, numTot);
						if (numTot > num)
							numTot = num;
						numRight = 0;
						numLeft = 0;
						numUp = 0;
						numDown = 0;
						
						if ((numTot == 0) || (num <= (4.0 / useDiffAll))) {
							//printf("num = %lf, matrix[%d][%d]=%d\n", num, i, j, matrix[i][j]);
							// its too small to take average
							// so randomize each one
							int k;
							
							for (k = 0; k < num; k++) {
								RND = (int) (RAND_NUM * (4.0 / useDiffAll));
								switch(RND) {
									case 0:
										++numRight;
										break;
									case 1:
										++numUp;
										break;
									case 2:
										++numLeft;
										break;
									case 3:
										++numDown;
										break;
								}
							}
						} else {
							// average the diffusion in all directions
							numRight = (int) floor((num * useDiffOne) + 0.5);
							numLeft = numRight;
							numUp = numRight;
							numDown = numRight;
							
							// rounding may cause the total to be
							// more than the actual number of molecules
							// that should be diffused
							if ((numRight * 4) > numTot) {
								do {
									// randomly decrease in one of
									// the directions until we reach
									// the right number of molecules
									RND = (int) (RAND_NUM * 4);
									switch(RND) {
										case 0:
											--numRight;
											if (numRight < 0) numRight = 0;
												break;
										case 1:
											--numLeft;
											if (numLeft < 0) numLeft = 0;
												break;
										case 2:
											--numUp;
											if (numUp < 0) numUp = 0;
												break;
										case 3:
											--numDown;
											if (numDown < 0) numDown = 0;
												break;
									}
								} while ((numRight + numLeft + numUp + numDown) > numTot);
							}
						}
						
						//printf("%d: numRight=%d, numUp=%d, numLeft=%d, numDown=%d\n", (*theGrid)[i][j],
						// numRight, numUp, numLeft, numDown);
						
						// diffuse the molecules
						// handle different boundary conditions
						(*theGrid)[i][j] -= numRight;
						if (j+1 == width) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[i][0] = (*theGrid)[i][0] + numRight;
							else
								(*theGrid)[i][j] += numRight;
						} else
							(*theGrid)[i][j+1] = (*theGrid)[i][j+1] + numRight;
						
						(*theGrid)[i][j] -= numLeft;
						if (j == 0) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[i][width-1] = (*theGrid)[i][width-1] + numLeft;
							else
								(*theGrid)[i][j] += numLeft;
						} else
							(*theGrid)[i][j-1] = (*theGrid)[i][j-1] + numLeft;
						
						(*theGrid)[i][j] -= numUp;
						if (i == 0) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[height-1][j] = (*theGrid)[height-1][j] + numUp;
							else
								(*theGrid)[i][j] += numUp;
						} else
							(*theGrid)[i-1][j] = (*theGrid)[i-1][j] + numUp;
						
						(*theGrid)[i][j] -= numDown;
						if (i+1 == height) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[0][j] = (*theGrid)[0][j] + numDown;
							else
								(*theGrid)[i][j] += numDown;
						} else
							(*theGrid)[i+1][j] = (*theGrid)[i+1][j] + numDown;
						
						//printf("%d\n", (*theGrid)[i][j]);
					}
				}
			}
		}
	}
}

- (void)performDoubleReaction
{
	// get activator/inhibitor grids
	NSDictionary *aGrid = [grid gridWithName: activatorName];
	NSDictionary *iGrid = [grid gridWithName: inhibitorName];
	NSNumber *n = [aGrid objectForKey: @"height"];
	int height = [n intValue];
	n = [aGrid objectForKey: @"width"];
	int width = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	double (*activatorGrid)[height][width] = [m getMatrix];
	m = [iGrid objectForKey: @"matrix"];
	double (*inhibitorGrid)[height][width] = [m getMatrix];
	
	// reaction parameters
	NSString *s = [simParameters objectForKey: @"activatorOnActivator"];
	double activatorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"activatorOnInhibitor"];
	double activatorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnActivator"];
	double inhibitorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnInhibitor"];
	double inhibitorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"maxActivatorProduce"];
	int maxActivatorProduce = [s intValue];
	s = [simParameters objectForKey: @"maxInhibitorProduce"];
	int maxInhibitorProduce = [s intValue];
	//printf("activator: %lf %lf %d\n", activatorOnActivator, activatorOnInhibitor, maxActivatorProduce);
	//printf("inhibitor: %lf %lf %d\n", inhibitorOnActivator, inhibitorOnInhibitor, maxInhibitorProduce);
	
	// setup delegate
	[delegate setupReaction: grid withActivatorGrid: aGrid withInhibitorGrid: iGrid];
	
	// perform reaction
	int i, j;
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			
			if (delegatePreReaction)
				if (![delegate preReaction])
					continue;
			
			// activator produce activator
			double activator = (double)(*activatorGrid)[i][j];
			double actProduce = activator * activatorOnActivator;
			if ((maxActivatorProduce != -1) && (actProduce > maxActivatorProduce))
				actProduce = maxActivatorProduce;
			
			// activator produce inhibitor
			double inhibitor = (double)(*inhibitorGrid)[i][j];
			double inhibProduce = activator * activatorOnInhibitor;
			if ((maxInhibitorProduce != -1) && (inhibProduce > maxInhibitorProduce))
				inhibProduce = maxInhibitorProduce;
			
			if (delegateProduction)
				if (![delegate doubleProductionOfActivator: &actProduce ofInhibitor: &inhibProduce atX: j atY: i])
					continue;
			
			// inhibitor cancel activator
			double actCancel = inhibitor * inhibitorOnActivator;
			if ((actCancel + activator) < 0)
				actCancel = -activator;
			double cancelActivator = -actCancel;
			
			// inhibitor decays
			double inhibDecay = actCancel * inhibitorOnInhibitor;
			
			if (delegateAfterCalculate)
				[delegate afterCalculate];
			
			// update activator/inhibitor for this pixel
			(*activatorGrid)[i][j] = (*activatorGrid)[i][j] + actProduce - cancelActivator;
			(*inhibitorGrid)[i][j] = (*inhibitorGrid)[i][j] + inhibProduce - inhibDecay;
			
			/*
			 if ((*activatorGrid)[i][j] > 0) {
				 printf("%d %d: %lf + %d - %d = %d\n", i, j, activator, actProduce, cancelActivator, (*activatorGrid)[i][j]);
				 printf("%d %d: %lf + %d - %d = %d\n", i, j, inhibitor, inhibProduce, inhibDecay, (*inhibitorGrid)[i][j]);
			 } */
			
			if (delegatePostReaction)
				[delegate postReaction];
		}
	}
	
}

- (void)performDoubleDiffusion
{
	// get activator/inhibitor grids
	NSDictionary *aGrid = [grid gridWithName: activatorName];
	NSDictionary *iGrid = [grid gridWithName: inhibitorName];
	NSNumber *n = [aGrid objectForKey: @"height"];
	int height = [n intValue];
	n = [aGrid objectForKey: @"width"];
	int width = [n intValue];
	
	// diffusion parameters
	NSString *s = [simParameters objectForKey: @"activatorDiffusion"];
	double activatorDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionOnFibronectin"];
	double activatorDiffusionOnFibronectin = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionTime"];
	int activatorDiffusionTime = [s intValue];
	s = [simParameters objectForKey: @"inhibitorDiffusion"];
	double inhibitorDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionTime"];
	int inhibitorDiffusionTime = [s intValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionOnFibronectin"];
	double inhibitorDiffusionOnFibronectin = [s doubleValue];
	BOOL use_alt = NO;
	//printf("%lf %lf %d %lf %d\n", activatorDiffusion, activatorDiffusionOnFibronectin,
	// activatorDiffusionTime, inhibitorDiffusion, inhibitorDiffusionTime);
	
	int i, j, k, RND;
	double numTot, numRight, numLeft, numUp, numDown;
	double num, useDiffAll, useDiffOne;
	double diffRateAll, diffRateOne;
	double altDiffRateAll, altDiffRateOne;
	double matrix[height][width];
	int molecule;
	int diffusionTime;
	int bCond;
	
	//printf("performDiffusion\n");
	
	// diffusion for both activator and inhibitor
	for (molecule = 0; molecule < 2; ++molecule) {
		double (*theGrid)[height][width];
		
		if (molecule == 0) {
			// activator
			Grid2DArray *m = [aGrid objectForKey: @"matrix"];
			theGrid = [m getMatrix];
			diffusionTime = activatorDiffusionTime;
			
			// equal rate of diffusion for each direction
			diffRateAll = activatorDiffusion;
			diffRateOne = activatorDiffusion / 4.0;
			// alternate diffusion rate on fibronectin
			altDiffRateAll = activatorDiffusionOnFibronectin;
			altDiffRateOne = activatorDiffusionOnFibronectin / 4.0;
			// boundary condition
			n = [aGrid objectForKey: @"boundaryCondition"];
			bCond = [n intValue];
		} else {
			// inhibitor
			Grid2DArray *m = [iGrid objectForKey: @"matrix"];
			theGrid = [m getMatrix];
			diffusionTime = inhibitorDiffusionTime;
			
			// equal rate of diffusion for each direction
			diffRateAll = inhibitorDiffusion;
			diffRateOne = inhibitorDiffusion / 4.0;
			// alternate diffusion rate on fibronectin
			altDiffRateAll = inhibitorDiffusionOnFibronectin;
			altDiffRateOne = inhibitorDiffusionOnFibronectin / 4.0;
			// boundary condition
			n = [iGrid objectForKey: @"boundaryCondition"];
			bCond = [n intValue];
		}
		
		// perform diffusion some number of times
		for (k = 0; k < diffusionTime; ++k) {
			
			// Make copy of current grid
			for (i = 0; i < height; i++)
				for (j = 0; j < width; j++)
					matrix[i][j] = (*theGrid)[i][j];
			
			// diffuse
			for (i = 0; i < height; i++) {
				for (j = 0; j < width; j++) {
					
					// equal rate of diffusion for each direction
					useDiffAll = diffRateAll;
					useDiffOne = diffRateOne;
					// alternate diffusion rate on fibronectin
					if (use_alt) {
						//if (FN[i][j] > 0) {
						useDiffAll = altDiffRateAll;
						useDiffOne = altDiffRateOne;
						//}
					}
					
					if (matrix[i][j] > 0) {
						// determine amount
						num = (double)matrix[i][j];
						numTot = (int) floor((num * useDiffAll) + 0.5);
						//printf("%f %d\n", num, numTot);
						if (numTot > num)
							numTot = num;
						numRight = 0;
						numLeft = 0;
						numUp = 0;
						numDown = 0;
						
							// average the diffusion in all directions
							numRight = num * useDiffOne;
							numLeft = numRight;
							numUp = numRight;
							numDown = numRight;
							
						
						//printf("%d: numRight=%d, numUp=%d, numLeft=%d, numDown=%d\n", (*theGrid)[i][j],
						// numRight, numUp, numLeft, numDown);
						
						// diffuse the molecules
						// handle different boundary conditions
						(*theGrid)[i][j] -= numRight;
						if (j+1 == width) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[i][0] = (*theGrid)[i][0] + numRight;
							else
								(*theGrid)[i][j] += numRight;
						} else
							(*theGrid)[i][j+1] = (*theGrid)[i][j+1] + numRight;
						
						(*theGrid)[i][j] -= numLeft;
						if (j == 0) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[i][width-1] = (*theGrid)[i][width-1] + numLeft;
							else
								(*theGrid)[i][j] += numLeft;
						} else
							(*theGrid)[i][j-1] = (*theGrid)[i][j-1] + numLeft;
						
						(*theGrid)[i][j] -= numUp;
						if (i == 0) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[height-1][j] = (*theGrid)[height-1][j] + numUp;
							else
								(*theGrid)[i][j] += numUp;
						} else
							(*theGrid)[i-1][j] = (*theGrid)[i-1][j] + numUp;
						
						(*theGrid)[i][j] -= numDown;
						if (i+1 == height) {
							// periodic can move but noflux does not
							if (bCond == PERIODIC_BOUNDARY)
								(*theGrid)[0][j] = (*theGrid)[0][j] + numDown;
							else
								(*theGrid)[i][j] += numDown;
						} else
							(*theGrid)[i+1][j] = (*theGrid)[i+1][j] + numDown;
						
						//printf("%d\n", (*theGrid)[i][j]);
					}
				}
			}
		}
	}
}

// finite difference
- (void)performReactionDiffusion
{
	// get activator/inhibitor grids
	NSDictionary *aGrid = [grid gridWithName: activatorName];
	NSDictionary *iGrid = [grid gridWithName: inhibitorName];
	NSNumber *n = [aGrid objectForKey: @"height"];
	int height = [n intValue];
	n = [aGrid objectForKey: @"width"];
	int width = [n intValue];
	Grid2DArray *m = [aGrid objectForKey: @"matrix"];
	double (*activatorGrid)[height][width] = [m getMatrix];
	m = [iGrid objectForKey: @"matrix"];
	double (*inhibitorGrid)[height][width] = [m getMatrix];
	
	// reaction parameters
	NSString *s = [simParameters objectForKey: @"activatorOnActivator"];
	double activatorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"activatorOnInhibitor"];
	double activatorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnActivator"];
	double inhibitorOnActivator = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorOnInhibitor"];
	double inhibitorOnInhibitor = [s doubleValue];
	s = [simParameters objectForKey: @"maxActivatorProduce"];
	int maxActivatorProduce = [s intValue];
	s = [simParameters objectForKey: @"maxInhibitorProduce"];
	int maxInhibitorProduce = [s intValue];
	//printf("activator: %lf %lf %d\n", activatorOnActivator, activatorOnInhibitor, maxActivatorProduce);
	//printf("inhibitor: %lf %lf %d\n", inhibitorOnActivator, inhibitorOnInhibitor, maxInhibitorProduce);

	// diffusion parameters
	s = [simParameters objectForKey: @"activatorDiffusion"];
	double activatorDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionOnFibronectin"];
	double activatorDiffusionOnFibronectin = [s doubleValue];
	s = [simParameters objectForKey: @"activatorDiffusionTime"];
	int activatorDiffusionTime = [s intValue];
	s = [simParameters objectForKey: @"inhibitorDiffusion"];
	double inhibitorDiffusion = [s doubleValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionTime"];
	int inhibitorDiffusionTime = [s intValue];
	s = [simParameters objectForKey: @"inhibitorDiffusionOnFibronectin"];
	double inhibitorDiffusionOnFibronectin = [s doubleValue];

	int i, j;
	double aMatrix[height][width];
	double iMatrix[height][width];

	double dI = inhibitorDiffusion * inhibitorDiffusionTime;
	double dA = activatorDiffusion * activatorDiffusionTime;
	double deltaTime = 1.0 / (4.0 * dI);

	// setup delegate
	[delegate setupReaction: grid withActivatorGrid: aGrid withInhibitorGrid: iGrid];

	// Make copy of current grid
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++) {
			aMatrix[i][j] = (*activatorGrid)[i][j];
			iMatrix[i][j] = (*inhibitorGrid)[i][j];
		}

	// explicit method, finite difference
	for (i = 0; i < height; i++) {
		for (j = 0; j < width; j++) {
			
			if (delegatePreReaction)
				if (![delegate preReaction])
					continue;

			// activator produce activator
			//double activator = (double)(*activatorGrid)[i][j];
			double activator = aMatrix[i][j];
			double actProduce = activator * activatorOnActivator;
			if ((maxActivatorProduce != -1) && (actProduce > maxActivatorProduce))
				actProduce = maxActivatorProduce;
			
			// activator produce inhibitor
			//double inhibitor = (double)(*inhibitorGrid)[i][j];
			double inhibitor = iMatrix[i][j];
			double inhibProduce = activator * activatorOnInhibitor;
			if ((maxInhibitorProduce != -1) && (inhibProduce > maxInhibitorProduce))
				inhibProduce = maxInhibitorProduce;
			
			if (delegateProduction)
				if (![delegate doubleProductionOfActivator: &actProduce ofInhibitor: &inhibProduce atX: j atY: i])
					continue;
			
			// inhibitor cancel activator
			double actCancel = inhibitor * inhibitorOnActivator;
			if ((actCancel + activator) < 0)
				actCancel = -activator;
//			double cancelActivator = -actCancel;
			
			// inhibitor decays
			double inhibDecay = -actCancel * inhibitorOnInhibitor;
			
			if (delegateAfterCalculate)
				[delegate afterCalculate];

			// diffusion term
			int up, down, left, right;
			if (i == 0) up = height - 1;
			else up = i - 1;
			if (i+1 == height) down = 0;
			else down = i + 1;
			if (j == 0) left = width - 1;
			else left = j - 1;
			if (j+1 == width) right = 0;
			else right = j + 1;

			double diffuseActivator = 0.25 * (aMatrix[up][j] + aMatrix[down][j] + aMatrix[i][left]
					+ aMatrix[i][right]) - aMatrix[i][j];
			diffuseActivator = activatorDiffusion * diffuseActivator;
			double deltaActivator = deltaTime * (actProduce + actCancel) + deltaTime * dA * diffuseActivator;

			double diffuseInhibitor = 0.25 * (iMatrix[up][j] + iMatrix[down][j] + iMatrix[i][left]
					+ iMatrix[i][right]) - iMatrix[i][j];
			diffuseInhibitor = inhibitorDiffusion * diffuseInhibitor;
			double deltaInhibitor = deltaTime * (inhibProduce + inhibDecay) + 0.25 * diffuseInhibitor;

			// update activator/inhibitor for this pixel
			(*activatorGrid)[i][j] = (*activatorGrid)[i][j] + deltaActivator;
			if ((*activatorGrid)[i][j] < 0) (*activatorGrid)[i][j] = 0;
			(*inhibitorGrid)[i][j] = (*inhibitorGrid)[i][j] + deltaInhibitor;
			if ((*inhibitorGrid)[i][j] < 0) (*inhibitorGrid)[i][j] = 0;
			
#if 0
			if ((*inhibitorGrid)[i][j] > 1) {
				printf("%d %d: %lf + %lf(%lf) = %lf\n", i, j, activator, deltaActivator, diffuseActivator, (*activatorGrid)[i][j]);
				printf("%d %d: %lf + %lf(%lf) = %lf\n", i, j, inhibitor, deltaInhibitor, diffuseInhibitor, (*inhibitorGrid)[i][j]);
				printf("%d %d: %lf\n", up, j, iMatrix[up][j]);
				printf("%d %d: %lf\n", down, j, iMatrix[down][j]);
				printf("%d %d: %lf\n", i, left, iMatrix[i][left]);
				printf("%d %d: %lf\n", i, right, iMatrix[i][right]);
			}
#endif

			if (delegatePostReaction)
				[delegate postReaction];
		}
	}
	
}

@end
