//
// cpmodel.c
//
// Chondrogenic patterning simulation
//
// Contact: Scott Christley <schristl@nd.edu>
//

//
// Simulation description--
//
// Cells, initially of size 7 pixels, are distributed uniformly across
// a 2D grid; then they are randomly permutated to come up with the
// initial cell distribution.
//

#include "stdio.h"
#include "stdlib.h"
#include "math.h"
#include "time.h"

//
// Simulation parameters
//

// random seed
#define START_RAND_SEED 12345
//#define START_RAND_SEED 325
#define RAND_NUM ((double)rand() / (double)RAND_MAX )

//# define MAX 2147483648.0   //The Maximum of random number+1 
// 2D grid size
#define Igrid 63         
#define Jgrid 63         

// number of cells
#define CELLS 400

// number of simulation steps
#define CYCLES 50
// output data every X cycles
#define OUTPUT_RATE 5

// ratio between molecule/cell levels
#define N 2 // 2

// activator produced by cells
#define CELL_ACT_PROB 1.0
#define CELL_ACT 1
#define CELL_ACT_MIN 0
#define FCELL_ACT_PROB 1.0
#define FCELL_ACT 1
#define FCELL_ACT_MIN 10
#define FCELL_ACT_TIME 250000
#define FCELL_INHIB 0

// activator diffusion, self-activation, inhibitor suppress
#define ACT_DIFF 0.16 // 0.8 // 0.165
#define FIBRO_ACT_DIFF 0.155
#define SELF_ACT .3356  // .2097 // 0.09
#define ACT_ON_INHIB 0.16 // 0.16
#define MAX_ACT 1000   // 55

// inhibitor diffusion, inhibit of activator
#define INHIB_DIFF 0.95  // 1.0
#define INHIB_ON_ACT -1.1 // -.1.1
#define INHIB_ON_INHIB -0.4615 // -0.7351
#define MAX_INHIB 1000   // 55

// fibronectin parameters
#define FIBRO_DIFF 0.8        // initial fibronectin diffusion
#define ACT_THRESHOLD 1500    // level of activator needed for fibronectin production
#define INHIB_THRESHOLD 40000000    // level of activator needed for fibronectin production
#define MEMORY_THRESHOLD 100000   // 
#define FIBRO_THRESHOLD 10    // level of fibronectin needed to make the cells round up
#define MAX_FIBRO 20          // max number of fibronectin molecules allowed per node
#define FIBRO_PRODUCTION 1.0  // probability of a cell producing fibronectin when threshold is met
#define FIBRO_CELL 1.0        // probability of a cell stay in a FN cell phase

// cell movement probabilities
#define CELL_MOVEMENT 25
#define CELL_KEEPSHAPE 0.9
#define CELL_PREST 0.9
#define CELL_PRIGHT 0.025
#define CELL_PUP 0.025
#define CELL_PLEFT 0.025
#define CELL_PDOWN 0.025
// cell movement probabilities on fibronectin
#define FCELL_MOVEMENT 5
#define FCELL_KEEPSHAPE 0.5
#define FCELL_PREST 0.6
#define FCELL_PRIGHT 0.1
#define FCELL_PUP 0.1
#define FCELL_PLEFT 0.1
#define FCELL_PDOWN 0.1

//
// global variables .. yuck!
//

FILE *fp_cell, *fp_morpha, *fp_morphb, *fp_FN, *fp_FNcell;  //file pointer
char filename_cell[]="output_cell.dat", filename_morpha[]="output_morpha.dat", filename_morphb[]="output_morphb.dat"; 
char filename_FN[]="output_FN.dat", filename_FNcell[]="output_FNcell.dat";
int grid[Igrid][Jgrid], FNcells[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], FNcell_mark[CELLS];
int FNactivator[Igrid][Jgrid];
    //  FNcell (FNcell_mark==1), free cell (FNcell_mark==0)
int morpha[Igrid*N][Jgrid*N], morphb[Igrid*N][Jgrid*N], morpha_mark[CELLS], upthreshold_a[Igrid*N][Jgrid*N]; 
int morphaMemory[CELLS];
   //  morpha_mark==1, the position of the cell occupied reaches the trreshold and begin to produce FN
int FN[Igrid*N][Jgrid*N], newFN[Igrid*N][Jgrid*N], FNproducingnodes[Igrid*N][Jgrid*N];

//
// function declarations
//

void printmatrix_cell(int[Igrid][Jgrid]);
void printmatrix_FNcell(int[Igrid][Jgrid]);
void printmatrix_cellMemory();
void printmatrix_morphb(int[Igrid*N][Jgrid*N]);
void printmatrix_morpha(int[Igrid*N][Jgrid*N]);
void printmatrix_FN(int[Igrid*N][Jgrid*N]);
void buildcell_7(int[Igrid][Jgrid], int[CELLS][7], int[CELLS][7], int, int, int);
void protrusion(int[Igrid][Jgrid], int[CELLS][7], int[CELLS][7], int, int, int, int);
void cellmove_7(int[Igrid][Jgrid], int[CELLS][7], int[CELLS][7], int, double, double, double, double, double);
int checknewcoor(int[Igrid][Jgrid], int[7], int[7], int[CELLS][7], int[CELLS][7], int);
void rewritecell(int[Igrid][Jgrid], int[7], int[7], int[CELLS][7], int[CELLS][7], int);
void changeshape_7(int[Igrid][Jgrid], int[CELLS][7], int[CELLS][7], int);
void activation(int[Igrid*N][Jgrid*N], int[Igrid][Jgrid], double, int);
void morphogendiff(int[Igrid*N][Jgrid*N], double, double, int);
void FNproduction (int[Igrid*N][Jgrid*N], double, int, double);
int check_FNcell(int, int);
int check_FNcell_5(int[5], int[5], int);
void change_7to5(int);
void cellmove_5(int[Igrid][Jgrid], int[Igrid][Jgrid], int[CELLS][7], int[CELLS][7], int, double,
		double, double, double, double, int);
int checknewcoor_5(int[Igrid][Jgrid], int[5], int[5], int[CELLS][7], int[CELLS][7], int);
void rewritecell_5(int[Igrid][Jgrid], int[Igrid][Jgrid], int[5], int[5], int[CELLS][7], int[CELLS][7], int);
void changeshape_5(int[Igrid][Jgrid], int[CELLS][7], int[CELLS][7], int);
int check_threshold_a(int[Igrid][Jgrid], int[Igrid*N][Jgrid*N], int, int[CELLS][7], int[CELLS][7], int, int);



/* printmatrix can print a matrix of (Igrid-2)*(Jgrid-2) */
void printmatrix_cell(Matrix_cell)
     int Matrix_cell[Igrid][Jgrid];
{
  int i,j;

  for(i=1;i<Igrid-1;i++)
    for(j=1;j<Jgrid-1;j++)
      if(fwrite(&Matrix_cell[i][j],sizeof(int),1,fp_cell)!=1)
	printf("file write error\n");
}

void printmatrix_FNcell(Matrix_FNcell)
     int Matrix_FNcell[Igrid][Jgrid];
{
  int i,j;

  for(i=1;i<Igrid-1;i++)
    for(j=1;j<Jgrid-1;j++)
      if(fwrite(&Matrix_FNcell[i][j],sizeof(int),1,fp_FNcell)!=1)
	printf("file write error\n");
}

void printmatrix_cellMemory()
{
  int i,j;
  int cellMemory[Igrid][Jgrid];

  for (i = 1; i < (Igrid - 1); i++)
    for (j = 1; j < (Jgrid - 1); j++)
      cellMemory[i][j] = 0;

  for (i = 0; i < CELLS; i++) {
    if(FNcell_mark[i] == 0) {
      for (j = 0; j < 7; j++) {
	cellMemory[A[i][j]][B[i][j]] = morphaMemory[i];
      }
    }
    if(FNcell_mark[i] == 1) {
      for (j = 0; j < 5; j++) {
	cellMemory[A[i][j]][B[i][j]] = morphaMemory[i];
      }
    }
  }

  for (i = 1; i < (Igrid - 1); i++)
    for (j = 1; j < (Jgrid - 1); j++)
      if(fwrite(&cellMemory[i][j],sizeof(int),1,fp_FNcell)!=1)
	printf("file write error\n");

}

void printmatrix_morpha(Matrix_morpha)
     int Matrix_morpha[Igrid*N][Jgrid*N];
{
  int i,j;

  for(i=1*N;i<(Igrid-1)*N;i++)
    for(j=1*N;j<(Jgrid-1)*N;j++)
      if(fwrite(&Matrix_morpha[i][j],sizeof(int),1,fp_morpha)!=1)
	printf("file write error\n");
}

void printmatrix_morphb(Matrix_morphb)
     int Matrix_morphb[Igrid*N][Jgrid*N];
{
  int i,j;

  for(i=1*N;i<(Igrid-1)*N;i++)
    for(j=1*N;j<(Jgrid-1)*N;j++)
      if(fwrite(&Matrix_morphb[i][j],sizeof(int),1,fp_morphb)!=1)
	printf("file write error\n");
}

void printmatrix_FN(Matrix_FN)
     int Matrix_FN[Igrid*N][Jgrid*N];
{
  int i,j;

  for(i=1*N;i<(Igrid-1)*N;i++)
    for(j=1*N;j<(Jgrid-1)*N;j++)
      if(fwrite(&Matrix_FN[i][j],sizeof(int),1,fp_FN)!=1)
	printf("file write error\n");
}


/* buildcell_7 can create a cell with 7 pixels */
void buildcell_7(grid,A,B,x,y,cellnumber)
     int grid[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], x, y, cellnumber;
{
  int a, b;   //k and K[4] record the protruding pixels.
  int i, g, h;
  double r1, r2;

  grid[x][y]=1;        
  A[cellnumber][0]=x;
  B[cellnumber][0]=y;
  grid[x][y+1]=1;
  A[cellnumber][1]=x;
  B[cellnumber][1]=y+1;
  grid[x+1][y]=1;
  A[cellnumber][2]=x+1;
  B[cellnumber][2]=y;
  grid[x+1][y+1]=1;
  A[cellnumber][3]=x+1;   
  B[cellnumber][3]=y+1;
  
  r1 = RAND_NUM;
     
  if(r1<1/4.0)  //There is no protruding pixle on the right side of 2*2 block.
    protrusion(grid,A,B,0,x,y,cellnumber);

  if(r1>=1/4.0 && r1<1/2.0)    //There is no protruding pixle on the up side of 2*2 block.
    protrusion(grid,A,B,1,x,y,cellnumber); 

  if(r1>=1/2.0 && r1<3/4.0)    //There is no protruding pixle on the left side of 2*2 block.
    protrusion(grid,A,B,2,x,y,cellnumber);

  if(r1>=3/4.0)    //There is no protruding pixle on the down side of 2*2 block.
    protrusion(grid,A,B,3,x,y,cellnumber);
}


/* protrusion can create 3 protrusion pixels on 3 sides of the 2*2 square */
void protrusion(grid,A,B,unaccept,x,y,cellnumber)
     int grid[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], unaccept ,x, y, cellnumber;
{
  int C[8],D[8];
  int pixels_pro;
  int i, temp0, temp1, temp2;
  double r2;

  C[0]=x+1;
  D[0]=y+2;
  C[1]=x;
  D[1]=y+2;
  C[2]=x-1;
  D[2]=y+1;
  C[3]=x-1;
  D[3]=y;
  C[4]=x;
  D[4]=y-1;
  C[5]=x+1;
  D[5]=y-1;
  C[6]=x+2;
  D[6]=y;
  C[7]=x+2;
  D[7]=y+1;  
  pixels_pro=0;

  for(i=0;i<4;i++)
    {
      if(i==unaccept) continue;

      temp0=i*2;
      temp1=i*2+1;
      temp2=pixels_pro+4;

      if(grid[C[temp0]][D[temp0]]==1)
	{
	  grid[C[temp1]][D[temp1]]=1;
	  A[cellnumber][temp2]=C[temp1];   
	  B[cellnumber][temp2]=D[temp1];
	  pixels_pro++;
	  continue;
	}
      if(grid[C[temp1]][D[temp1]]==1)
	{
	  grid[C[temp0]][D[temp0]]=1;
	  A[cellnumber][temp2]=C[temp0];   
	  B[cellnumber][temp2]=D[temp0];
	  pixels_pro++;
	  continue;
	}

      r2 = RAND_NUM;
      if(r2<1/2.0)
	{ 
	  grid[C[temp0]][D[temp0]]=1;
	  A[cellnumber][temp2]=C[temp0];   
	  B[cellnumber][temp2]=D[temp0];
	}
      else   //r2>=1/2.0
	{ 
	  grid[C[temp1]][D[temp1]]=1;
	  A[cellnumber][temp2]=C[temp1];   
	  B[cellnumber][temp2]=D[temp1];
	}
      pixels_pro++;
    }
}




/* cellmove_7 can make the cell move as a whole with a certain probability. The cell don't change shape. */
void cellmove_7(grid,A,B,cellnumber,Prest,Pright,Pup,Pleft,Pdown)
     int grid[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], cellnumber;
     double Prest,Pright,Pup,Pleft,Pdown;
{
  double r;
  int i;
  int Anew[7],Bnew[7];
  int check;

  r = RAND_NUM;
  /*  printf("r=%lf\n",r); */
  if(r<Pright)   // Move to the right.
    {
      for(i=0;i<7;i++)
	{
	  /*  printf("when r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew[i]=A[cellnumber][i];
	  Bnew[i]=B[cellnumber][i]+1;
          if(Bnew[i]==Jgrid-1)   Bnew[i]=1;    //periodic boundary condition
	  /*  printf("when r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove,i,Anew[i],i,Bnew[i]); */
	}
      check=checknewcoor(grid,Anew,Bnew,A,B,cellnumber);
      /*  printf("check=%d\n",check); */
      if(check==1)
	rewritecell(grid,Anew,Bnew,A,B,cellnumber);
    }
  
  if(r>=Pright && r<Pright+Pup)   // Go up.
    {
      for(i=0;i<7;i++)
	{
	  /*  printf("when  %lf<r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove,Pmove*2,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew[i]=A[cellnumber][i]-1;
          if(Anew[i]==0)   Anew[i]=Igrid-2;    //periodic boundary condition
	  Bnew[i]=B[cellnumber][i];
	  /* printf("when %lf<r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove,Pmove*2,i,Anew[i],i,Bnew[i]); */
	}
      check=checknewcoor(grid,Anew,Bnew,A,B,cellnumber);
      /*  printf("check=%d\n",check); */
      if(check==1)
	rewritecell(grid,Anew,Bnew,A,B,cellnumber);
    }

  if(r>=Pright+Pup && r<Pright+Pup+Pleft)   // Move to the left.
    {
      for(i=0;i<7;i++)
	{
	  /*  printf("when  %lf<r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove*2,Pmove*3,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew[i]=A[cellnumber][i];
	  Bnew[i]=B[cellnumber][i]-1;
          if(Bnew[i]==0)   Bnew[i]=Jgrid-2;    //periodic boundary condition
	  /*  printf("when %lf<r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove*2,Pmove*3,i,Anew[i],i,Bnew[i]); */
	}
      check=checknewcoor(grid,Anew,Bnew,A,B,cellnumber);
      /* printf("check=%d\n",check); */
      if(check==1)
	rewritecell(grid,Anew,Bnew,A,B,cellnumber);
    }
  
  if(r>=Pright+Pup+Pleft && r<Pright+Pup+Pleft+Pdown)   // Go down.
    {
      for(i=0;i<7;i++)
	{
	  /* printf("when  %lf<r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove*3,Pmove*4,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew[i]=A[cellnumber][i]+1;
          if(Anew[i]==Igrid-1)   Anew[i]=1;    //periodic boundary condition
	  Bnew[i]=B[cellnumber][i];
	  /* printf("when %lf<r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove*3,Pmove*4,i,Anew[i],i,Bnew[i]) */;
	}
      check=checknewcoor(grid,Anew,Bnew,A,B,cellnumber);
      /* printf("check=%d\n",check); */
      if(check==1)
	rewritecell(grid,Anew,Bnew,A,B,cellnumber);
    }
  /*  if(r>=4*Pmove)    printf("REST\n"); */
}



/* checknewcoor can check if the cells overlap.  */
/* At first, we assume a cell move in some direction, then use this function to check whether the pixels overlap. */
/* If they do, it returns "0" and we can not accept the movement. */
/* If they don't, it returns "1", and we can accept the movement. */  

int checknewcoor(grid,Anew,Bnew,A,B,cellnumber)   //The 1st sub-function of cellmove.
     int grid[Igrid][Jgrid], Anew[7], Bnew[7], A[CELLS][7], B[CELLS][7], cellnumber;
{
  int i, j;
  int Diff[7];  //Mark of the difference between A,B and Anew,Bnew.

  for(i=0;i<7;i++)   Diff[i]=1;

  for(i=0;i<7;i++)
    {
      for(j=0;j<7;j++)
	{
	  if(Anew[i]==A[cellnumber][j] && Bnew[i]==B[cellnumber][j]) 
	    {
	      /* printf("i=%d    j=%d\n",i,j); */
	      Diff[i]=0;
	      /* printf("Anew[%d]=%d    Bnew[%d]=%d\n",i,Anew[i],i,Bnew[i]); */
	    }
	}
    }
  
  for(i=0;i<7;i++)
    {
      if(Diff[i]==1)
	{
	  if(grid[Anew[i]][Bnew[i]]>0)
	    {
	      /* printf("grid[Anew[%d]][Bnew[%d]]==%d\n",i,i,grid[Anew[i]][Bnew[i]]); */
	      return(0);
	    }
	}
    }
  return(1);
}


/* If last function "checknewcoor" returns "1",  */
/* we accept the movement and use rewritecell to record to new position of the pixels  */
/* in the grid and their coordinates in A and B.  */
void rewritecell(grid,Anew,Bnew,A,B,cellnumber)   // The 2nd sub-function of cellmove.
     int grid[Igrid][Jgrid], Anew[7], Bnew[7], A[CELLS][7], B[CELLS][7], cellnumber;
{
  int i;

  for(i=0;i<7;i++)
    {
      grid[A[cellnumber][i]][B[cellnumber][i]]=0;
      /* printf("grid[A[%d][%d]][B[%d][%d]]=%d\n",cellnumber,i,cellnumber,i,grid[A[cellnumber][i]][B[cellnumber][i]]); */
    }

  for(i=0;i<7;i++)
    {
      grid[Anew[i]][Bnew[i]]=1;
      /* printf("grid[Anew[%d]][Bnew[%d]]=%d\n",i,i,grid[Anew[i]][Bnew[i]]); */
    }

  for(i=0;i<7;i++)
    {
      A[cellnumber][i]=Anew[i];
      B[cellnumber][i]=Bnew[i];
      /* printf("A[%d][%d]=%d    B[%d][%d]=%d\n",cellnumber,i,A[cellnumber][i],cellnumber,i, B[cellnumber][i]); */
    }
}



/* changeshape_7 can let the 7 pixel cells change their shapes freely. The pixels also can not overlap. */
void changeshape_7(grid,A,B,cellnumber)
     int grid[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], cellnumber;
{
  int i,j;
  int gridtest[4][4];
  int x0,y0;
  int side1[4],side2[4],sum_side1;
  int C[8],D[8],E[8],F[8];
  int pro_pixels;
  double P;
  double r1,r2;

  x0=A[cellnumber][0];
  y0=B[cellnumber][0];
  //printf("x0=%d    y0=%d   cellnumber=%d\n", x0, y0, cellnumber);

  for(i=0;i<4;i++)
    {
      side1[i]=0;
      side2[i]=0;
      //printf("side1[%d]=%d    side2[%d]=%d\n",i,side1[i],i,side2[i]);
    }

  for(i=0;i<4;i++)
    for(j=0;j<4;j++)
      gridtest[i][j]=0;

  
  C[0]=x0+1;
  D[0]=y0+2;
  C[1]=x0;
  D[1]=y0+2;
  C[2]=x0-1;
  D[2]=y0+1;
  C[3]=x0-1;
  D[3]=y0;
  C[4]=x0;
  D[4]=y0-1;
  C[5]=x0+1;
  D[5]=y0-1;
  C[6]=x0+2;
  D[6]=y0;
  C[7]=x0+2;
  D[7]=y0+1;  

  //Periodic boundary condition
  for(i=0;i<8;i++)
    {
      if(C[i]==0)         C[i]=Igrid-2;  
      if(C[i]==Igrid-1)   C[i]=1; 
      if(C[i]==Igrid)     C[i]=2;
      if(D[i]==0)         D[i]=Jgrid-2;  
      if(D[i]==Jgrid-1)   D[i]=1; 
      if(D[i]==Igrid)     D[i]=2;
    }

  E[0]=2;      
  F[0]=3;
  E[1]=1;
  F[1]=3;
  E[2]=0;
  F[2]=2;
  E[3]=0;
  F[3]=1;
  E[4]=1;
  F[4]=0;
  E[5]=2;
  F[5]=0;
  E[6]=3;
  F[6]=1;
  E[7]=3;
  F[7]=2; 

  //  for(i=0;i<8;i++) 
  //    { 
  //     if(C[i]==0 || C[i]>61 || D[i]==0 || D[i]>61)
  //        printf("C[%d]=%d    D[%d]=%d\n",i,C[i],i,D[i]); 
  //    }  

  /*
  for(i=0;i<8;i++) 
    { 
      printf("C[%d]=%d    D[%d]=%d    E[%d]=%d    F[%d]=%d\n",i,C[i],i,D[i],i,E[i],i,F[i]); 
    } 
  */

  for(i=4;i<7;i++)
    grid[A[cellnumber][i]][B[cellnumber][i]]=0;

  for(i=0;i<8;i++)
    gridtest[E[i]][F[i]]=grid[C[i]][D[i]];

  /*  for(i=0;i<4;i++) */
  /*     { */
  /*       for(j=0;j<4;j++) */
  /* 	printf("%d",gridtest[i][j]);  */
  /*       printf("\n"); */
  /*     }  */

  for(i=0;i<4;i++)
    {
      if(gridtest[E[i*2]][F[i*2]]==0 || gridtest[E[i*2+1]][F[i*2+1]]==0)  
	{
	  side1[i]=1;
	  side2[i]=1;
	}
      if(gridtest[E[i*2]][F[i*2]]==0 && gridtest[E[i*2+1]][F[i*2+1]]==0)  
	side2[i]=2;

      //printf("gridtest[%d][%d]=%d     gridtest[%d][%d]=%d\n",E[i*2],F[i*2],
      //     gridtest[E[i*2]][F[i*2]],E[i*2+1],F[i*2+1],gridtest[E[i*2+1]][F[i*2+1]]);
      //printf("side1[%d]=%d    side2[%d]=%d\n",i,side1[i],i,side2[i]);
    }

  sum_side1=0;
  for(i=0;i<4;i++)
    {
      sum_side1=sum_side1+side1[i];
      /*   printf("sum_side1=%d\n",sum_side1); */
    }
  if(sum_side1<3)
    {
      printf("Debug1!\n");
    }
   
  P=0.25;
  pro_pixels=0;
  r1 = RAND_NUM;
  /*  printf("r1=%lf\n",r1);   */
  
  if(sum_side1==4)
    {
      for(i=0;i<4;i++)
	if(r1>=i*P && r1<(i+1)*P)   side1[i]=0;
    }
  
  for(i=0;i<4;i++)
    {
      if(side1[i]==0)   continue;
      if(side2[i]==2)
	{
	  r2 = RAND_NUM;
	  /*  printf("r2=%lf\n",r2); */
	  if (r2<0.5)    
	    {
	      grid[C[i*2]][D[i*2]]=1;
	      A[cellnumber][pro_pixels+4]=C[i*2];
	      B[cellnumber][pro_pixels+4]=D[i*2];
	    }
	  else  // r2>=0.5
	    {
	      grid[C[i*2+1]][D[i*2+1]]=1;
	      A[cellnumber][pro_pixels+4]=C[i*2+1];
	      B[cellnumber][pro_pixels+4]=D[i*2+1];
	    } 
	}
      else   //  side2[i]==1
	{
	  if(gridtest[E[i*2]][F[i*2]]==0)   
	    {
	      grid[C[i*2]][D[i*2]]=1;
	      A[cellnumber][pro_pixels+4]=C[i*2];
	      B[cellnumber][pro_pixels+4]=D[i*2];
	    }
	  else  
	    {
	      grid[C[i*2+1]][D[i*2+1]]=1;
	      A[cellnumber][pro_pixels+4]=C[i*2+1];
	      B[cellnumber][pro_pixels+4]=D[i*2+1];
	    } 
	}
      pro_pixels++;
    }
  
}


/* activation can calculate the interaction between activator and inhibitor */
void activation (morphogen,cellgrid,effect,max)
     int morphogen[Igrid*N][Jgrid*N],cellgrid[Igrid][Jgrid],max;
     double effect;
{
  int i,j,k1,k2;
  int t1;
  int production[Igrid*N][Jgrid*N], activator[Igrid*N][Jgrid*N], maximumproduction[Igrid*N][Jgrid*N];
  double x;

  for(i=1;i<Igrid-1;i++)
    {
      for(j=1;j<Jgrid-1;j++)
	{
	  if(cellgrid[i][j]==2)
	    for(k1=0;k1<N;k1++)
	      for(k2=0;k2<N;k2++)
		{
		  activator[i*N+k1][j*N+k2]=morpha[i*N+k1][j*N+k2]*cellgrid[i][j]/2;
		}
	  else
	    for(k1=0;k1<N;k1++)
	      for(k2=0;k2<N;k2++)
		{
		  activator[i*N+k1][j*N+k2]=morpha[i*N+k1][j*N+k2]*cellgrid[i][j];
		}
	  for(k1=0;k1<N;k1++)
	    for(k2=0;k2<N;k2++)
	      {
		maximumproduction[i*N+k1][j*N+k2]=max*cellgrid[i][j];
	      }
	}
    }

  for(i=1*N;i<(Igrid-1)*N;i++)
    {
      for(j=1*N;j<(Jgrid-1)*N;j++)
	{
	  x=effect*activator[i][j];
	  t1=floor(x);
	  if((x-t1)<0.5)
	    production[i][j]=t1;
	  else
	    production[i][j]=t1+1;
	  
	}
    }
  
  for(i=1*N;i<(Igrid-1)*N;i++)
    {
      for(j=1*N;j<(Jgrid-1)*N;j++)
	{
	  if(production[i][j]>maximumproduction[i][j])
	    production[i][j]=maximumproduction[i][j];
	}
    }

  for(i=1*N;i<(Igrid-1)*N;i++)
    {
      for(j=1*N;j<(Jgrid-1)*N;j++)
	{
	  morphogen[i][j]=morphogen[i][j]+production[i][j];
	  if(morphogen[i][j]<0)
	    morphogen[i][j]=0;
	}
    }

}

// Perform diffusion of chemicals
void morphogendiff(morphogen, diff, alt_diff, use_alt)
     int morphogen[Igrid*N][Jgrid*N];
     double diff, alt_diff;
     int use_alt;
{
  int i, j, RND;
  int numTot, numRight, numLeft, numUp, numDown;
  double num, diff1, alt_diff1, use_diff, use_diff1;
  int morphogen0[Igrid*N][Jgrid*N];

  // equal rate of diffusion
  // for each direction
  diff1 = diff/4.0;
  alt_diff1 = alt_diff/4.0;
  //printf("diff=%f %f\n", diff, diff1);

  // Make copy of current grid
  for(i = N; i < (Igrid-1)*N; i++)
    for(j = N; j < (Jgrid-1)*N; j++)
      morphogen0[i][j]=morphogen[i][j];

  // diffuse
  for (i = N; i < (Igrid-1)*N; i++) {
    for (j = N; j < (Jgrid-1)*N; j++) {

      // alternate diffusion rate on fibronectin
      use_diff = diff;
      use_diff1 = diff1;
      if (use_alt) {
	if (FN[i][j] > 0) {
	  use_diff = alt_diff;
	  use_diff1 = alt_diff1;
	}
      }

      if (morphogen0[i][j] > 0) {
	// determine amount
	num = (double)morphogen0[i][j];
	numTot = (int) floor((num * use_diff) + 0.5);
	//printf("%f %d\n", num, numTot);
	if (numTot > num)
	  numTot = num;
	numRight = 0;
	numLeft = 0;
	numUp = 0;
	numDown = 0;

	if ((numTot == 0) || (num <= (4.0 / use_diff))) {
	  //printf("num = %lf, morphogen0[%d][%d]=%d\n", num, i, j, morphogen0[i][j]);
	  // its too small to take average
	  // so randomize each one
	  int k;
	  double diff2 = use_diff1 * 2;
	  double diff3 = use_diff1 * 3;
	  double diff4 = use_diff1 * 4;

	  for (k = 0; k < num; k++) {
	    RND = (int) (RAND_NUM * (4.0 / use_diff));
	    switch(RND) {
	    case 0:
	      ++numRight;
	      break;
	    case 1:
	      ++numUp;
	      break;
	    case 2:
	      ++numLeft;
	      break;
	    case 3:
	      ++numDown;
	      break;
	    }
	  }
	  //printf("numRight=%d, numUp=%d, numLeft=%d, numDown=%d\n", numRight, numUp, numLeft, numDown);
	} else {
	  // average the diffusion in all directions
	  numRight = (int) floor((num * use_diff1) + 0.5);
	  numLeft = numRight;
	  numUp = numRight;
	  numDown = numRight;

	  // rounding may cause the total to be
	  // more than the actual number of molecules
	  // that should be diffused
	  if ((numRight * 4) > numTot) {
	    do {
	      // randomly decrease in one of
	      // the directions until we reach
	      // the right number of molecules
	      RND = (int) (RAND_NUM * 4);
	      switch(RND) {
	      case 0:
		--numRight;
		if (numRight < 0) numRight = 0;
		break;
	      case 1:
		--numLeft;
		if (numLeft < 0) numLeft = 0;
		break;
	      case 2:
		--numUp;
		if (numUp < 0) numUp = 0;
		break;
	      case 3:
		--numDown;
		if (numDown < 0) numDown = 0;
		break;
	      }
	    } while ((numRight + numLeft + numUp + numDown) > numTot);
	  }
	}

	// diffuse the molecules
	morphogen[i][j] -= numRight;
	if (j+1 == N*(Jgrid-1))
	  morphogen[i][N] = morphogen[i][N] + numRight;
	else
	  morphogen[i][j+1] = morphogen[i][j+1] + numRight;

	morphogen[i][j] -= numLeft;
	if (j-1 == N-1)
	  morphogen[i][N*(Jgrid-1)-1] = morphogen[i][N*(Jgrid-1)-1] + numLeft;
	else
	  morphogen[i][j-1] = morphogen[i][j-1] + numLeft;

	morphogen[i][j] -= numUp;
	if (i-1 == N-1)
	  morphogen[N*(Igrid-1)-1][j] = morphogen[N*(Igrid-1)-1][j] + numUp;
	else
	  morphogen[i-1][j] = morphogen[i-1][j] + numUp;

	morphogen[i][j] -= numDown;
	if (i+1 == N*(Igrid-1))
	  morphogen[N][j] = morphogen[N][j] + numDown;
	else
	  morphogen[i+1][j] = morphogen[i+1][j] + numDown;
      }
    }
  }
}

/* In morphogendiff, the particals diffuse with a certain probability and the unit is one pixel. */
void old_morphogendiff(morphogen,diff)
     int morphogen[Igrid*N][Jgrid*N];
     double diff;
{
  int i, j, k, num;
  double RND, diff1;
  int morphogen0[Igrid*N][Jgrid*N];

  /*  printf("Test morphogendiff beginning\n"); */

  diff1=diff/4.0;

  for(i=N;i<(Igrid-1)*N;i++)
    for(j=N;j<(Jgrid-1)*N;j++)
      morphogen0[i][j]=morphogen[i][j];

  for(i=N;i<(Igrid-1)*N;i++)
    {
      for(j=N;j<(Jgrid-1)*N;j++)
	{
	  if(morphogen0[i][j]>0)
	    {
	      num=morphogen0[i][j];
	      for(k=0;k<num;k++)
		{
		  RND = RAND_NUM;
		  if(RND<=diff1)    //right
		    {
		      morphogen[i][j]=morphogen[i][j]-1;
                      if(j+1==N*(Jgrid-1))  //periodic boundary condition
                        morphogen[i][N]=morphogen[i][N]+1;
                      else
		        morphogen[i][j+1]=morphogen[i][j+1]+1;
		    }
		  
		  if(RND>diff1 && RND<=diff1*2)  //up
		    {
		      morphogen[i][j]=morphogen[i][j]-1;
                      if(i-1==N-1)  //periodic boundary condition
                        morphogen[N*(Igrid-1)-1][j]=morphogen[N*(Igrid-1)-1][j]+1;
                      else
		        morphogen[i-1][j]=morphogen[i-1][j]+1;
		    }
		  
		  if(RND>diff1*2 && RND<=diff1*3)   //left
		    {
		      morphogen[i][j]=morphogen[i][j]-1;
                      if(j-1==N-1)  //periodic boundary condition
                        morphogen[i][N*(Jgrid-1)-1]=morphogen[i][N*(Jgrid-1)-1]+1;
                      else
		        morphogen[i][j-1]=morphogen[i][j-1]+1;
		    }
		  
		  if(RND>diff1*3 && RND<=diff1*4)   //down
		    {
		      morphogen[i][j]=morphogen[i][j]-1;
                      if(i+1==N*(Igrid-1))  //periodic boundary condition
                        morphogen[N][j]=morphogen[N][j]+1;
                      else
		        morphogen[i+1][j]=morphogen[i+1][j]+1;
		    }
		}
	    }
	}
    }

  /*  printf("Test morphogendiff end\n"); */

}


/* FNproduction  */
/* this function creates a fibronectin in a neighborhood of radius 1 with probability p */
/* wherever there are cells at a fibronectin-producing node */
/* a node is fibronectin producing for 600 timesteps */
/* whenever the threshold t in morphgen concentration m is met */
void FNproduction (upthreshold_a, production, mf, diff_FN)
     int upthreshold_a[Igrid*N][Jgrid*N], mf;
     double production, diff_FN;
{
  int i, j, k1, k2, tempx, tempy;
  double r;

  /*   printf("FNproducingnodes\n"); */
  /*   printmatrix(FNproducingnodes); */
  /*   printf("newFN\n"); */
  /*   printmatrix(newFN); */
  /*  STEP1: Update fibronectin-producing nodes. */
  /*   printmatrix(morpha); */
  /* decrease all non-zero fibronectin-production timers of all nodes by 1 */
  /*  printf("TEST0\n"); */
  for(i=1*N;i<(Igrid-1)*N;i++)
    {
      for(j=1*N;j<(Jgrid-1)*N;j++)
	/* printf("TEST1%d][%d]\n",i,j); */
	if(FNproducingnodes[i][j]>0)
	  {
	    FNproducingnodes[i][j]=FNproducingnodes[i][j]-1;	   
	  }
    } 

  /*   account for threshold_a */
  /* for(i=1;i<Igrid-1;i++) */
  /*     { */
  /*       for(j=1;j<Jgrid-1;j++) */
  /* 	if(morpha[i][j]>=threshold_a) */
  /* 	  thresh_nodes[i][j]=1; */
  /* 	else */
  /* 	  thresh_nodes[i][j]=0; */
  /*     } */

  /* nodes with threshold fibronectin will produce fibronectin for at least another 600 timesteps */
  /*  printf("newFN\n"); */
  /*   printmatrix(newFN); */
  /*   printf("FNproducingnodes\n"); */
  /*   printmatrix(FNproducingnodes); */
  for(i=1;i<Igrid-1;i++)
    {
      for(j=1;j<Jgrid-1;j++)
	{
	  /*  printf("TEST2[%d][%d]\n",i,j); */
	  for(k1=0;k1<N;k1++)
	    for(k2=0;k2<N;k2++)
	      {
		if(upthreshold_a[i*N+k1][j*N+k2]>0)
		  {
		    FNproducingnodes[i*N+k1][j*N+k2]=FNproducingnodes[i*N+k1][j*N+k2]+600;

		    //printf("threshold_a=%d\n",threshold_a);
		    //printf("morpha[%d][%d]=%d\n",i,j,morpha[i][j]);
		    //printf("FNproducingnodes[%d][%d]=%d\n",i,j,FNproducingnodes[i][j]);
		  }
	 
		/*  printf("threshold_a=%d\n",threshold_a); */
		if(FNproducingnodes[i*N+k1][j*N+k2]>0)
		  {
		    /*  STEP2: Begin producing fibronectin */
		    /* fibronectin is deposited with probability production */

		    r = RAND_NUM;
		    /*  printf("r=%lf\n",r); */
		    /* 	      printf("production=%lf\n",production); */
		    if(r<=production)
		      {
			if(grid[i][j]==2)
			  newFN[i*N+k1][j*N+k2]=grid[i][j]/2;
			else
			  newFN[i*N+k1][j*N+k2]=grid[i][j];
		 
			//printf("grid[%d][%d]=%d\n",i,j,grid[i][j]);
			//printf("newFN[%d][%d]=%d\n", i*N+k1, j*N+k2, newFN[i*N+k1][j*N+k2]);
		      }
		    if(FNproducingnodes[i*N+k1][j*N+k2]>600)
		      FNproducingnodes[i*N+k1][j*N+k2]=600;
		  }
		else
		  {
		    /* fproduction[i][j]=0; */
		    /*  fplaces[i][j]=0; */
		    newFN[i*N+k1][j*N+k2]=0;
		    /*  printf("newFN[%d][%d]=%d\n",i,j,newFN[i][j]); */
		  }
		
	      }
	}
    }

  /*  STEP2: Begin producing fibronectin */

  /* fibronectin is deposited with probability production */
  /*  for(i=1;i<Igrid-1;i++) */
  /*     { */
  /*       for(j=1;j<Jgrid-1;j++) */
  /* 	{ */
  /* 	  r=rand()/MAX; */
  /* 	  if(r<=production) */
  /* 	    newFN[i][j]=fplaces[i][j]; */
  /* 	} */
  /*     } */


  /*   the actual deposition can be anywhere within one node */
  morphogendiff(newFN, diff_FN, diff_FN, 0);

  /* then add on old fibronectin */
  for(i=1*N;i<(Igrid-1)*N;i++)
    {
      for(j=1*N;j<(Jgrid-1)*N;j++)
	{
	  FN[i][j]= FN[i][j]+newFN[i][j];
	  if(FN[i][j]>mf)
	    FN[i][j]=mf;  //allowing only mf fibronectins per node
	}
    }

}




/* check_FNcell can determine whether the cell is in the fibronectin domain. */
/* For a 7 pixel cell,it is FNcell if 4 or more pixels is in the fibronectin domain */
int check_FNcell(cellnumber,threshold_FN)
     int cellnumber, threshold_FN;
{
  int i, k1, k2, count_FN;
  int halfcover;
  double x;

  x=7.0*N*N/2.0;
  halfcover=floor(x)+1;
  
  count_FN=0;
  for(i=0;i<7;i++)
    { 
      for(k1=0;k1<N;k1++)
	for(k2=0;k2<N;k2++)
	  {
	    if(FN[A[cellnumber][i]*N+k1][B[cellnumber][i]*N+k2]>=threshold_FN) {
	      count_FN++;
	      //printf("FN[%d][%d]=%d\n", A[cellnumber][i]*N+k1, B[cellnumber][i]*N+k2,
	      //FN[A[cellnumber][i]*N+k1][B[cellnumber][i]*N+k2]);
	    }
	    if(count_FN>=halfcover)
	      {
		/* printf("TEST:check_FNcell result=1\n"); */
		return(1);
	      }
	  }
    }
  /* printf("TEST:check_FNcell result=0\n"); */
  return(0);
}


/* check_FNcell can determine whether the cell is in the fibronectin domain. */
/* For a 5 pixel cell,it is FNcell if 3 or more pixels is in the fibronectin domain */
int check_FNcell_5(Anew_5,Bnew_5,threshold_FN)
     int threshold_FN, Anew_5[5], Bnew_5[5];
{
  int i, k1, k2, count_FN;
  int halfcover;
  double x;

  x=5.0*N*N/2.0;
  halfcover=floor(x)+1;
  
  count_FN=0;
  for(i=0;i<5;i++)
    { 
      for(k1=0;k1<N;k1++)
	for(k2=0;k2<N;k2++)
	  {
	    if(FN[Anew_5[i]*N+k1][Bnew_5[i]*N+k2]>=threshold_FN)
	      count_FN++;
	    if(count_FN>=halfcover)
	      {
		/* printf("TEST:check_FNcell result=1\n"); */
		return(1);
	      }
	  }
    }
  /* printf("TEST:check_FNcell result=0\n"); */
  return(0);
}



/* check_freecell can determine whether the cell goes out of the fibronectin domain. */
/* For a 5 pixel cell,it is FNcell if 3 or more pixels is out of the fibronectin domain */
/* check_freecell(cellnumber,threshold_FN) */
/* int cellnumber; */
/* { */
/*   int i,count_free; */
  
/*   count_free=0; */
/*   for(i=0;i<5;i++) */
/*     { */
/*       if(FN[A[cellnumber][i]][B[cellnumber][i]]<threshold_FN) */
/* 	count_free++; */
/*       if(count_free==3) */
/* 	return(0); */
/*     } */
/*   return(1); */
/* } */



/* change_7to5 can change a cell with 7 pixels to 5 pixels*/
void change_7to5(cellnumber)
     int cellnumber;
{
  int k;
  double r;
  
  r = RAND_NUM;
     
  if(r<=1/3.0)  //Keep A[][4],B[][4]
    {
      /* shrink(grid,A,B,cellnumber,5); */
      /*       shrink(grid,A,B,cellnumber,6); */
      grid[A[cellnumber][5]][B[cellnumber][5]]=0;
      grid[A[cellnumber][6]][B[cellnumber][6]]=0;
      for(k=0;k<5;k++)
	grid[A[cellnumber][k]][B[cellnumber][k]]=2;
     
    }

  if(r>1/3.0 && r<=2/3.0)    //Keep 5
    {
      /* shrink(grid,A,B,cellnumber,4); */
      /*       shrink(grid,A,B,cellnumber,6); */
      grid[A[cellnumber][4]][B[cellnumber][4]]=0;
      grid[A[cellnumber][6]][B[cellnumber][6]]=0;
      A[cellnumber][4]=A[cellnumber][5];
      B[cellnumber][4]=B[cellnumber][5];
      for(k=0;k<5;k++)
	grid[A[cellnumber][k]][B[cellnumber][k]]=2;
     
     
    }

  if(r>2/3.0)    //Keep 6
    {
      /*  shrink(grid,A,B,cellnumber,4); */
      /*       shrink(grid,A,B,cellnumber,5); */
      grid[A[cellnumber][4]][B[cellnumber][4]]=0;
      grid[A[cellnumber][5]][B[cellnumber][5]]=0;
      A[cellnumber][4]=A[cellnumber][6];
      B[cellnumber][4]=B[cellnumber][6];
      for(k=0;k<5;k++)
	grid[A[cellnumber][k]][B[cellnumber][k]]=2;
      
      
    }
  A[cellnumber][5]=0;
  B[cellnumber][5]=0;
  A[cellnumber][6]=0;
  B[cellnumber][6]=0;

  printf("TEST:change_7to5\n");
}







/* cellmove_5 can make the cell move as a whole with a certain probability. The cell don't change shape. */
void cellmove_5(grid,FNcells,A,B,cellnumber,Prest,Pright,Pup,Pleft,Pdown,threshold_FN)
     int grid[Igrid][Jgrid], FNcells[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], cellnumber, threshold_FN;
     double Prest,Pright,Pup,Pleft,Pdown;
{
  double r;
  int i;
  int Anew_5[5],Bnew_5[5];
  int check,check2;

  r = RAND_NUM;
  /*  printf("r=%lf\n",r); */
  if(r<Pright)   // Move to the right.
    {
      for(i=0;i<5;i++)
	{
	  /*  printf("when r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew_5[i]=A[cellnumber][i];
	  Bnew_5[i]=B[cellnumber][i]+1;
          if(Bnew_5[i]==Jgrid-1)   Bnew_5[i]=1;    //periodic boundary condition
	  /*  printf("when r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove,i,Anew[i],i,Bnew[i]); */
	}
      check=checknewcoor_5(grid,Anew_5,Bnew_5,A,B,cellnumber);
      /*  printf("checkright=%d\n",check); */
      if(check==1)
	{
	  check2=check_FNcell_5(Anew_5,Bnew_5,threshold_FN);
	  if(check2==1)
	    rewritecell_5(grid,FNcells,Anew_5,Bnew_5,A,B,cellnumber);
  	}
    }
  
  if(r>=Pright && r<Pright+Pup)   // Go up.
    {
      for(i=0;i<5;i++)
	{
	  /*  printf("when  %lf<r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove,Pmove*2,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew_5[i]=A[cellnumber][i]-1;
          if(Anew_5[i]==0)   Anew_5[i]=Igrid-2;    //periodic boundary condition
	  Bnew_5[i]=B[cellnumber][i];
	  /* printf("when %lf<r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove,Pmove*2,i,Anew[i],i,Bnew[i]); */
	}
      check=checknewcoor_5(grid,Anew_5,Bnew_5,A,B,cellnumber);
      /*  printf("checkup=%d\n",check); */
      if(check==1)
	{
	  check2=check_FNcell_5(Anew_5,Bnew_5,threshold_FN);
	  if(check2==1)
	    rewritecell_5(grid,FNcells,Anew_5,Bnew_5,A,B,cellnumber);
  	}
    }

  if(r>=Pright+Pup && r<Pright+Pup+Pleft)   // Move to the left.
    {
      for(i=0;i<5;i++)
	{
	  /*  printf("when  %lf<r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove*2,Pmove*3,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew_5[i]=A[cellnumber][i];
	  Bnew_5[i]=B[cellnumber][i]-1;
          if(Bnew_5[i]==0)   Bnew_5[i]=Jgrid-2;    //periodic boundary condition
	  /*  printf("when %lf<r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove*2,Pmove*3,i,Anew[i],i,Bnew[i]); */
	}
      check=checknewcoor_5(grid,Anew_5,Bnew_5,A,B,cellnumber);
      /*  printf("checkleft=%d\n",check); */
      if(check==1)
	{
	  check2=check_FNcell_5(Anew_5,Bnew_5,threshold_FN);
	  if(check2==1)
	    rewritecell_5(grid,FNcells,Anew_5,Bnew_5,A,B,cellnumber);
  	}
    }
  
  if(r>=Pright+Pup+Pleft && r<Pright+Pup+Pleft+Pdown)   // Go down.
    {
      for(i=0;i<5;i++)
	{
	  /* printf("when  %lf<r<%lf, A[%d][%d]=%d     B[%d][%d]=%d\n",Pmove*3,Pmove*4,cellnumber,i,A[cellnumber][i],cellnumber,i,B[cellnumber][i]); */
	  Anew_5[i]=A[cellnumber][i]+1;
          if(Anew_5[i]==Igrid-1)   Anew_5[i]=1;    //periodic boundary condition
	  Bnew_5[i]=B[cellnumber][i];
	  /* printf("when %lf<r<%lf, Anew[%d]=%d     Bnew[%d]=%d\n",Pmove*3,Pmove*4,i,Anew[i],i,Bnew[i]) */;
	}
      check=checknewcoor_5(grid,Anew_5,Bnew_5,A,B,cellnumber);
      /*  printf("checkdown=%d\n",check); */
      if(check==1)
	{
	  check2=check_FNcell_5(Anew_5,Bnew_5,threshold_FN);
	  if(check2==1)
	    rewritecell_5(grid,FNcells,Anew_5,Bnew_5,A,B,cellnumber);
  	}
    }
  /*  if(r>=4*Pmove)    printf("REST\n"); */
  /*  printf("TEST:cellmove_5\n"); */

}



/* checknewcoor_5 can check if the cells overlap.  */
/* At first, we assume a cell move in some direction, then use this function to check whether the pixels overlap. */
/* If they do, it returns "0" and we can not accept the movement. */
/* If they don't, it returns "1", and we can accept the movement. */  

int checknewcoor_5(grid,Anew_5,Bnew_5,A,B,cellnumber)   //The 1st sub-function of cellmove.
     int grid[Igrid][Jgrid], Anew_5[5], Bnew_5[5], A[CELLS][7], B[CELLS][7], cellnumber;
{
  int i, j;
  int Diff_5[5];  //Mark of the difference between A,B and Anew_5,Bnew_5.

  for(i=0;i<5;i++)   Diff_5[i]=1;

  for(i=0;i<5;i++)
    {
      for(j=0;j<5;j++)
	{
	  if(Anew_5[i]==A[cellnumber][j] && Bnew_5[i]==B[cellnumber][j]) 
	    {
	      /* printf("i=%d    j=%d\n",i,j); */
	      Diff_5[i]=0;
	      /* printf("Anew[%d]=%d    Bnew[%d]=%d\n",i,Anew[i],i,Bnew[i]); */
	    }
	}
    }
  
  for(i=0;i<5;i++)
    {
      if(Diff_5[i]==1)
	{
	  if(grid[Anew_5[i]][Bnew_5[i]]>0)
	    {
	      /* printf("grid[Anew[%d]][Bnew[%d]]==%d\n",i,i,grid[Anew[i]][Bnew[i]]); */
	      /* printf("TEST:checknewcoor_5 result=0\n"); */
	      return(0);
	    }
	}
    }
  /* printf("TEST:checknewcoor_5 result=1\n"); */
  return(1);
}


/* If last function "checknewcoor" returns "1",  */
/* we accept the movement and use rewritecell to record to new position of the pixels  */
/* in the grid and their coordinates in A and B.  */
void rewritecell_5(grid,FNcells,Anew_5,Bnew_5,A,B,cellnumber)   // The 2nd sub-function of cellmove.
     int grid[Igrid][Jgrid],FNcells[Igrid][Jgrid], Anew_5[5], Bnew_5[5], A[CELLS][7], B[CELLS][7], cellnumber;
{
  int i;
  int save[5];

  for(i=0;i<5;i++)
    {
      save[i]=grid[A[cellnumber][i]][B[cellnumber][i]];
      grid[A[cellnumber][i]][B[cellnumber][i]]=0;
      FNcells[A[cellnumber][i]][B[cellnumber][i]]=0;
      /* printf("grid[A[%d][%d]][B[%d][%d]]=%d\n",cellnumber,i,cellnumber,i,grid[A[cellnumber][i]][B[cellnumber][i]]); */
    }

  for(i=0;i<5;i++)
    {
      grid[Anew_5[i]][Bnew_5[i]]=save[i];
      FNcells[Anew_5[i]][Bnew_5[i]]=save[i];
      /* printf("grid[Anew[%d]][Bnew[%d]]=%d\n",i,i,grid[Anew[i]][Bnew[i]]); */
    }

  for(i=0;i<5;i++)
    {
      A[cellnumber][i]=Anew_5[i];
      B[cellnumber][i]=Bnew_5[i];
      /* printf("A[%d][%d]=%d    B[%d][%d]=%d\n",cellnumber,i,A[cellnumber][i],cellnumber,i, B[cellnumber][i]); */
    }
  /*  printf("TEST:rewritecell_5\n"); */
}



/* changeshape_5 can let the 5 pixel cells change their shapes freely. The pixels also can not overlap. */
void changeshape_5(grid,A,B,cellnumber)
     int grid[Igrid][Jgrid], A[CELLS][7], B[CELLS][7], cellnumber;
{
  int i,j,g;
  int gridtest[4][4];
  int x0,y0;
  int side1[4],side2[4],sum_side1;
  int C[8],D[8],E[8],F[8];
  double P;
  double r1,r2;

  x0=A[cellnumber][0];
  y0=B[cellnumber][0];
  //printf("x0=%d    y0=%d   cellnumber=%d\n", x0, y0, cellnumber);

  for(i=0;i<4;i++)
    {
      side1[i]=0;
      side2[i]=0;
      //printf("side1[%d]=%d    side2[%d]=%d\n",i,side1[i],i,side2[i]);
    }

  for(i=0;i<4;i++)
    for(j=0;j<4;j++)
      gridtest[i][j]=0;

  /*  for(i=0;i<4;i++) */
  /*     { */
  /*       for(j=0;j<4;j++) */
  /* 	printf("%d",gridtest[i][j]);  */
  /*       printf("\n"); */
  /*     }  */

  
  C[0]=x0+1;
  D[0]=y0+2;
  C[1]=x0;
  D[1]=y0+2;
  C[2]=x0-1;
  D[2]=y0+1;
  C[3]=x0-1;
  D[3]=y0;
  C[4]=x0;
  D[4]=y0-1;
  C[5]=x0+1;
  D[5]=y0-1;
  C[6]=x0+2;
  D[6]=y0;
  C[7]=x0+2;
  D[7]=y0+1;

  //Periodic boundary condition
  for(i=0;i<8;i++)
    {
      if(C[i]==0)         C[i]=Igrid-2;  
      if(C[i]==Igrid-1)   C[i]=1; 
      if(C[i]==Igrid)     C[i]=2;
      if(D[i]==0)         D[i]=Jgrid-2;  
      if(D[i]==Jgrid-1)   D[i]=1; 
      if(D[i]==Igrid)     D[i]=2;
    }

  E[0]=2;      
  F[0]=3;
  E[1]=1;
  F[1]=3;
  E[2]=0;
  F[2]=2;
  E[3]=0;
  F[3]=1;
  E[4]=1;
  F[4]=0;
  E[5]=2;
  F[5]=0;
  E[6]=3;
  F[6]=1;
  E[7]=3;
  F[7]=2;  

  //  for(i=0;i<8;i++) 
  //    { 
  //     if(C[i]==0 || C[i]>61 || D[i]==0 || D[i]>61)
  //        printf("C[%d]=%d    D[%d]=%d\n",i,C[i],i,D[i]); 
  //    }  

  //for(i = 0; i < 8; i++) {
  //printf("C[%d]=%d    D[%d]=%d    E[%d]=%d    F[%d]=%d\n",i,C[i],i,D[i],i,E[i],i,F[i]);
  //}

  //  for (i = 0; i < 8; ++i) {
  //    grid[A[cellnumber][i]][B[cellnumber][i]]=0;
  //    FNcells[A[cellnumber][i]][B[cellnumber][i]]=0;
  //  }

  for(i=0;i<8;i++)
    gridtest[E[i]][F[i]]=grid[C[i]][D[i]];

  /*
  for(i = 0; i < 4; i++) {
    for(j = 0; j < 4; j++)
      printf("%d",gridtest[i][j]);
    printf("\n");
  }
  */

  for(i=0;i<4;i++)
    {
      if(gridtest[E[i*2]][F[i*2]]==0 || gridtest[E[i*2+1]][F[i*2+1]]==0)  
	{
	  side1[i]=1;
	  side2[i]=1;
	}
      if(gridtest[E[i*2]][F[i*2]]==0 && gridtest[E[i*2+1]][F[i*2+1]]==0)  
	side2[i]=2;

      /*  printf("gridtest[%d][%d]=%d     gridtest[%d][%d]=%d\n",E[i*2],F[i*2],gridtest[E[i*2]][F[i*2]],E[i*2+1],F[i*2+1],gridtest[E[i*2+1]][F[i*2+1]]); */
      /*       printf("side1[%d]=%d    side2[%d]=%d\n",i,side1[i],i,side2[i]); */
    }
  
  sum_side1=0;
  for(i=0;i<4;i++)
    {
      sum_side1=sum_side1+side1[i];
      /*   printf("sum_side1=%d\n",sum_side1); */
    }
  
  if(sum_side1==0)
    {
      //printf("Debug2!\n");
    }
  
  P=1.0/sum_side1;
  r1 = RAND_NUM;
  /*  printf("r1=%lf\n",r1);   */
  
  g=0;
  for(i=0;i<4;i++)
    {
      if(side1[i]==1)
	{  
	  if(r1>=g*P && r1<(g+1)*P)
	    {
	      if(side2[i]==2)
		{
		  r2 = RAND_NUM;
		  /*  printf("r2=%lf\n",r2); */
		  if (r2<0.5)    
		    {
		      grid[C[i*2]][D[i*2]]=2;
		      FNcells[C[i*2]][D[i*2]]=2;
		      A[cellnumber][4]=C[i*2];
		      B[cellnumber][4]=D[i*2];
		    }
		  else  // r2>=0.5
		    {
		      grid[C[i*2+1]][D[i*2+1]]=2;
		      FNcells[C[i*2+1]][D[i*2+1]]=2;
		      A[cellnumber][4]=C[i*2+1];
		      B[cellnumber][4]=D[i*2+1];
		    } 
		}
	      else   //  side2[i]==1
		{
		  if(gridtest[E[i*2]][F[i*2]]==0)   
		    {
		      grid[C[i*2]][D[i*2]]=2;
		      FNcells[C[i*2]][D[i*2]]=2;
		      A[cellnumber][4]=C[i*2];
		      B[cellnumber][4]=D[i*2];
		    }
		  else  
		    {
		      grid[C[i*2+1]][D[i*2+1]]=2;
		      FNcells[C[i*2+1]][D[i*2+1]]=2;
		      A[cellnumber][4]=C[i*2+1];
		      B[cellnumber][4]=D[i*2+1];
		    } 
		}
	      i=5;  //jump out of the "for" circle
	    }
	  else
	    g++;

	} //end if(side1[i]==1)
    } //end for
  /*  printf("TEST:changeshape_5\n"); */
}

  

int newcheck_threshold_a(grid, morpha, cellnumber, A, B, threshold_a, cell_num_mark)
     int grid[Igrid][Jgrid], morpha[Igrid*N][Jgrid*N], A[CELLS][7], B[CELLS][7], cell_num_mark;
     int cellnumber, threshold_a;
{
  int t, t1, t2;
  int Ia_coor, Ja_coor;
  int found = 0;

  for (t = 0; t < cell_num_mark; ++t) {
    for (t1 = 0; t1 < N; ++t1) {
      for (t2 = 0; t2 < N; ++t2) {
	Ia_coor=A[cellnumber][t]*N+t1;
	Ja_coor=B[cellnumber][t]*N+t2;
	if(morpha[Ia_coor][Ja_coor]>=threshold_a)
	  found = 1;
      }
    }
  }

  return found;
}

int check_threshold_a(grid, morpha, cellnumber, A, B, threshold_a, cell_num_mark)
     int grid[Igrid][Jgrid], morpha[Igrid*N][Jgrid*N], A[CELLS][7], B[CELLS][7], cell_num_mark;
     int cellnumber, threshold_a;
{
  int t, t1, t2;
  int Ia_coor, Ja_coor;
  int found = 0;

  //morphaMemory[cellnumber] = 0;

  for (t = 0; t < cell_num_mark; ++t) {
    for (t1 = 0; t1 < N; ++t1) {
      for (t2 = 0; t2 < N; ++t2) {
	Ia_coor=A[cellnumber][t]*N+t1;
	Ja_coor=B[cellnumber][t]*N+t2;

	if (morpha[Ia_coor][Ja_coor] >= threshold_a)
	  found = 1;

	if (FNcell_mark[cellnumber] == 0) {
	  if (morpha[Ia_coor][Ja_coor] >= MEMORY_THRESHOLD)
	    morphaMemory[cellnumber] += MEMORY_THRESHOLD;
	  else
	    morphaMemory[cellnumber] += morpha[Ia_coor][Ja_coor];
	}

      }
    }
  }

  return found;
}

//Another way to check threshold_a according to Audi's suggestion
int old_check_threshold_a(grid, morpha, cellnumber, A, B, threshold_a, cell_num_mark)
     int grid[Igrid][Jgrid], morpha[Igrid*N][Jgrid*N], A[CELLS][7], B[CELLS][7], cell_num_mark;
     int cellnumber, threshold_a;
{
  double r, r1, r2;
  double z, z1, z2;
  int t, t1, t2;
  int Ia_coor, Ja_coor;

  r = RAND_NUM;
  z=r*cell_num_mark;
  t=floor(z); 

  r1 = RAND_NUM;
  z1=r1*N;
  t1=floor(z1); 

  r2 = RAND_NUM;
  z2=r2*N;
  t2=floor(z2); 

  Ia_coor=A[cellnumber][t]*N+t1;
  Ja_coor=B[cellnumber][t]*N+t2;
  //if(Ia_coor==(Igrid-1)*N)     Ia_coor=N;
  //if(Ia_coor==(Igrid-1)*N+1)   Ia_coor=N+1;
  //if(Ja_coor==(Jgrid-1)*N)     Ja_coor=N;
  //if(Ja_coor==(Jgrid-1)*N+1)   Ja_coor=N+1;
  if(morpha[Ia_coor][Ja_coor]>=threshold_a)
    return(1);
  else
    return(0);
}

void runSimulation()
{
  int i, j, k1, k2, cycle;
  int x_coor, y_coor;
  int cellnumber, ith;
  int tempx, tempy;
  double r, r0;
  int save_a;
  int savemorpha[Igrid*N][Jgrid*N];
  int savemorphb[Igrid*N][Jgrid*N];
  int t, t1, t2;
  double z, z1, z2, zd;
  int a_perpixel;
  int COUNT_FNcells;
 
  /*  open the binary output files */
  if((fp_cell=fopen(filename_cell,"wb+"))==NULL)
    {
      printf("cannot open file\n");
      exit(0);
    }

  if((fp_morpha=fopen(filename_morpha,"wb+"))==NULL)
    {
      printf("cannot open file\n");
      exit(0);
    }

  if((fp_morphb=fopen(filename_morphb,"wb+"))==NULL)
    {
      printf("cannot open file\n");
      exit(0);
    }
  
  if((fp_FN=fopen(filename_FN,"wb+"))==NULL)
    {
      printf("cannot open file\n");
      exit(0);
    }

  if((fp_FNcell=fopen(filename_FNcell,"wb+"))==NULL)
    {
      printf("cannot open file\n");
      exit(0);
    }
    


  /* fibronectin parameters */
  for(i=0;i<CELLS;i++)
    {
      FNcell_mark[i]=0;  // They are all free cells at the beginning.
      morpha_mark[i]=0;  // There is no morpha at the beginning.
      morphaMemory[i] = 0;
    }

  COUNT_FNcells=0;

  
  /* Initialize the grid of cells, morphgena, morphgenb and fibronectin population*/
  tempx=Igrid-1;
  tempy=Jgrid-1;

  for(i=1;i<=Igrid-2;i++)
    for(j=1;j<=Jgrid-2;j++)
      {
	grid[i][j]=0;
	FNcells[i][j]=0;
	FNactivator[i][j] = 0;

	for(k1=0;k1<N;k1++)
	  for(k2=0;k2<N;k2++)
	    {
	      morpha[i*N+k1][j*N+k2]=0;
	      morphb[i*N+k1][j*N+k2]=0;
	      FN[i*N+k1][j*N+k2]=0;	
	      FNproducingnodes[i*N+k1][j*N+k2]=0;
	      newFN[i*N+k1][j*N+k2]=0;
	      upthreshold_a[i*N+k1][j*N+k2]=0;
	    }
      }


  // Initialize RNG
  srand(START_RAND_SEED);

  /*
  int numRight = 0, numUp = 0, numLeft = 0, numDown = 0;
  for (i = 0; i < 1000000; ++i) {
    int RND = (int) (RAND_NUM * 4);
    switch(RND) {
    case 0:
      ++numRight;
      break;
    case 1:
      ++numUp;
      break;
    case 2:
      ++numLeft;
      break;
    case 3:
      ++numDown;
      break;
    default:
      printf("RND=%d\n", RND);
    }
  }
  printf("numRight=%d numUp=%d numLeft=%d numDown=%d\n", numRight, numUp, numLeft, numDown);
  */

  /* Initial Distribution of Cells*/
  // assume square grid
  cellnumber=0;
  int range = (int) sqrt(CELLS);
  int bound = Igrid / range;
  printf("range=%d\n",range);
  
  for(i = 0; i < range; i++)
    for(j = 0; j < range; j++)
      {
	x_coor = 2 + bound * i;
	y_coor = 2 + bound * j;
	buildcell_7(grid,A,B,x_coor,y_coor,cellnumber);
	cellnumber++;
      }
  printf("cellnumber=%d\n",cellnumber);
    
  /*  printf("CELLGRID\n"); */
  printmatrix_cell(grid); 

  for(i=0;i<100;i++)
    {
      for(j=0;j<CELLS;j++)
	{
	  r = RAND_NUM;
	  z = r * CELLS;
	  t=floor(z);
          changeshape_7(grid,A,B,t);
	  cellmove_7(grid, A, B, t, 0.2, 0.2, 0.2, 0.2, 0.2);
	}
      /*  printmatrix(grid); */
    }
  /*  printf("CELLGRID\n"); */
  printmatrix_cell(grid);

  // main simulation loop
  for(cycle = 0; cycle < CYCLES; cycle++)
    {
      printf("cycle=%d\n",cycle);
      printf("COUNT_FNcells=%d\n",COUNT_FNcells);

      //if(COUNT_FNcells>150)  break; 
     
      /* printf("cycle=%d\n",cycle); */
     

      // The following is to set the mark matrix equal to zero.
      for(i=0;i<CELLS;i++)
	{
	  if(morpha_mark[i]==1)
	    {
	      if(FNcell_mark[i]==0)
		for(j=0;j<7;j++)
		  for(k1=0;k1<N;k1++)
		    for(k2=0;k2<N;k2++)
		      upthreshold_a[A[i][j]*N+k1][B[i][j]*N+k2]=0;  
	      
	      else                     //(FNcell_mark[i]==1)
		for(j=0;j<5;j++)
		  for(k1=0;k1<N;k1++)
		    for(k2=0;k2<N;k2++)
		      upthreshold_a[A[i][j]*N+k1][B[i][j]*N+k2]=0;  
	    }
	}

      for(i=0;i<Igrid*N;i++)
	{
	  for(j=0;j<Jgrid*N;j++)
	    {
	      savemorpha[i][j]=morpha[i][j];
	      savemorphb[i][j]=morphb[i][j];
	    }
	}

      /*  cells diffusion and shape change */
      for (ith = 0; ith < CELLS; ith++) {
	// normal cells
	if (FNcell_mark[ith] == 0) {
	  if ((cycle % CELL_MOVEMENT) == 0) {
	    r0 = RAND_NUM;
	    if(r0 < (1 - CELL_KEEPSHAPE))
	      changeshape_7(grid, A, B, ith);
	    cellmove_7(grid, A, B, ith, CELL_PREST, CELL_PRIGHT, CELL_PUP,
		       CELL_PLEFT, CELL_PDOWN);
	  }
	}

	// fibronectin cells
	if (FNcell_mark[ith] == 1) {
	  if ((cycle % FCELL_MOVEMENT) == 0) {
	    r0 = RAND_NUM;
	    if(r0 < (1 - FCELL_KEEPSHAPE))
	      changeshape_5(grid, A, B, ith);
	    cellmove_5(grid, FNcells, A, B, ith, FCELL_PREST, FCELL_PRIGHT, FCELL_PUP,
		       FCELL_PLEFT, FCELL_PDOWN, FIBRO_THRESHOLD); 
	  }	 
	}
      }
      
      
    

      
      /*  activation only occurs where cells are, proportional to the concentration of morphogena */

      // activation of a upon b
      activation (morphb,grid, ACT_ON_INHIB, MAX_INHIB);

      // activation of a upon itself
      activation (morpha,grid, SELF_ACT, MAX_ACT);


      // cells can randomly produce activator
      for(i=1;i<Igrid-1;i++)
	for(j=1;j<Jgrid-1;j++)
	  for(k1=0;k1<N;k1++)
	    for(k2=0;k2<N;k2++)
              {
		r = RAND_NUM;
		/*
		if (r < CELL_ACT_PROB)
		  a_perpixel = CELL_ACT;
		else
		  a_perpixel = CELL_ACT_MIN;
		*/

		if(FNcells[i][j] == 0) {
		  if (r < CELL_ACT_PROB)
		    a_perpixel = CELL_ACT;
		  else
		    a_perpixel = CELL_ACT_MIN;
		} else {
		  if (FNactivator[i][j] > 0) {
		    if (r < FCELL_ACT_PROB)
		      a_perpixel = FCELL_ACT;
		    else
		      a_perpixel = FCELL_ACT_MIN;
		    --FNactivator[i][j];
		  } else {
		    // produce inhibitor?
		    a_perpixel = FCELL_INHIB;
		    morphb[i*N+k1][j*N+k2]=morphb[i*N+k1][j*N+k2]+grid[i][j]*a_perpixel; 
		  }
		}


		//z=r*2;
		//a_perpixel=floor(z)*49+1;  
		//printf("a_perpixel=%d\n",a_perpixel);

		/*
		int hold = morpha[i*N+k1][j*N+k2] - savemorpha[i*N+k1][j*N+k2];
		if (hold + a_perpixel > MAX_ACT)
		  a_perpixel = MAX_ACT - hold;
		*/

		morpha[i*N+k1][j*N+k2]=morpha[i*N+k1][j*N+k2]+grid[i][j]*a_perpixel; 

		// 1/(N*N) cell pixel can produce a_perpixel morphogena
              }



      // Morphb will affect morpha and itself.  
      for(i=1*N;i<(Igrid-1)*N;i++)
	for(j=1*N;j<(Jgrid-1)*N;j++)
	  {
	    save_a = morpha[i][j];
	    //save_a = savemorpha[i][j];

	    //if ((cycle % 10) == 0)
	    //printf("%d (%d %d): %d %d -> %d %d -> ", grid[i][j], i, j, savemorpha[i][j], savemorphb[i][j], morpha[i][j], morphb[i][j]);

	    /*   morphogenb effects morphogena */
	    z1 = INHIB_ON_ACT * savemorphb[i][j] + morpha[i][j];
	    //zd = INHIB_ON_ACT * savemorphb[i][j];
	    //z1 = morpha[i][j] + zd;
	    if(z1 > 0)
	      {
		/*  printf("z1=%lf\n",z1); */
		t1=floor(z1);
		/*  printf("t1=%d\n",t1); */
		if((z1-t1)<0.5)
		  morpha[i][j] = t1;
		else
		  morpha[i][j] = t1+1;
	      }
	    else
	      {
		morpha[i][j] = 0;
		z1 = 0;
		zd = 0;
		/*   printf("z1=%lf\n",z1); */
	      }
	    /* morphgenb decays as it inhibits morphogena */
	    //z2 = morphb[i][j] + (save_a - z1) / INHIB_ON_ACT; 
	    //z2 = morphb[i][j] + (save_a - z1) * INHIB_ON_ACT; 
	    z2 = morphb[i][j] + (save_a - z1) * INHIB_ON_INHIB; 
	    //z2 = morphb[i][j] + zd; 
	    /*  printf("%lf=%d+(%d-%lf)/%lf\n",z2,morphb[i][j],save_a,z1,beffecta); */
	    /*   printf("z2=%lf\n",z2); */
	    t2=floor(z2);
	    /*  printf("t2=%d\n",t2); */
	    if((z2-t2)<0.5)
	      morphb[i][j]=t2;
	    else
	      morphb[i][j]=t2+1;

	    //if ((cycle % 10) == 0)
	    //printf("%d %d : %f %f %f\n", morpha[i][j], morphb[i][j], zd, z1, z2);
	  }


      //morphogen diffusion
      //for(i=0;i<10;i++)
      for(i = 0; i < 100; i++)
        {
          morphogendiff(morpha, ACT_DIFF, FIBRO_ACT_DIFF, 1);
        }
      for(i = 0; i < 100; i++)
        {
	  morphogendiff(morphb, INHIB_DIFF, INHIB_DIFF, 0);
        }
      
      for(i=0;i<CELLS;i++)
	{
	  if(FNcell_mark[i]==0)
	    {
	      //morpha_mark[i]=check_threshold_a(grid, morphb, i, A, B, INHIB_THRESHOLD, 7);
	      morpha_mark[i]=check_threshold_a(grid, morpha, i, A, B, ACT_THRESHOLD, 7);
	      if(morpha_mark[i]==1)
		{
		  //printf("morpha_mark[%d]=%d\n",i,morpha_mark[i]);
		  for(j=0;j<7;j++)
		    for(k1=0;k1<N;k1++)
		      for(k2=0;k2<N;k2++)
			upthreshold_a[A[i][j]*N+k1][B[i][j]*N+k2]=1;
		}
	    }

	  if(FNcell_mark[i]==1)
	    {
	      //morpha_mark[i]=check_threshold_a(grid, morphb, i, A, B, INHIB_THRESHOLD, 5);
	      morpha_mark[i]=check_threshold_a(grid, morpha, i, A, B, ACT_THRESHOLD, 5);
	      if(morpha_mark[i]==1)
		{
		  //printf("morpha_mark1[%d]=%d\n",i,morpha_mark[i]);
		  for(j=0;j<5;j++)
		    for(k1=0;k1<N;k1++)
		      for(k2=0;k2<N;k2++)
			upthreshold_a[A[i][j]*N+k1][B[i][j]*N+k2]=1;
		}
	    }
	}


      /*  fibronectin production for 600 timesteps if threshold_a is met */
      FNproduction(upthreshold_a, FIBRO_PRODUCTION, MAX_FIBRO, FIBRO_DIFF);
      

      for(i=0;i<CELLS;i++)
	{
	  if(FNcell_mark[i]==0)
	    {
	      /* printf("Test0\n"); */
	      FNcell_mark[i]=check_FNcell(i, FIBRO_THRESHOLD);
	      if(FNcell_mark[i]==1)
		{
		  COUNT_FNcells++;
		  change_7to5(i);
		  for (j = 0; j < 5; j++) {
		    FNcells[A[i][j]][B[i][j]]=grid[A[i][j]][B[i][j]]; //to create FNcells matrix
		    FNactivator[A[i][j]][B[i][j]] = FCELL_ACT_TIME;
		  }
		}
	    }

	  /*  if(FNcell_mark[i]==1) */
	  /* 	    FNcell_mark[i]=check_freecell(i,threshold_FN); */

	}


      if ((cycle % OUTPUT_RATE) == 0) {
	/*  printf("CELLGRID\n"); */
	printmatrix_cell(grid);
	/*  printf("MORPH A\n"); */
	printmatrix_morpha(morpha);
	/*       printf("MORPH B\n"); */
	printmatrix_morphb(morphb);
	/*    printf("FN\n"); */
	printmatrix_FN(FN);      
	//printmatrix_FNcell(FNcells); 
	printmatrix_cellMemory(); 
      }

    } 

  fclose(fp_cell);
  fclose(fp_morpha);
  fclose(fp_morphb);
  fclose(fp_FN);
  fclose(fp_FNcell);

}
